import cv2
import random
import numpy as np
import tensorflow as tf
from qoalai.networks.inception_resnetv2 import *
from comdutils.file_utils import get_filenames


class DeepLab(InceptionV4ResnetV2):
    def __init__(self, num_classes, 
                       input_tensor = None,
                       input_shape = (None, 300, 300, 3),
                       output_stride = 16,
                       is_training = True):
        """[summary]
        
        Arguments:
            num_classes {[type]} -- [description]
        
        Keyword Arguments:
            input_shape {tuple} -- [description] (default: {(None, 300, 300, 3)})
            base_architecture {str} -- [description] (default: {'resnet_v2_101'})
            output_stride {int} -- [description] (default: {16})
            learning_rate {float} -- [description] (default: {0.0001})
            is_training {bool} -- [description] (default: {True})
        """
        # ---------------------------------- #
        # class properties                   #
        # ---------------------------------- #
        self.num_classes = num_classes
        self.input_height = input_shape[1]
        self.input_width = input_shape[2]
        super(DeepLab, self).__init__(is_training = is_training)  
        
        if input_tensor is None:
            self.input = tf.placeholder(shape=input_shape, dtype=tf.float32)
        else:
            self.input = input_tensor
        
        if self.num_classes == 1:
            output_shape = (None, self.input_height, self.input_width, 1)
        else:
            output_shape = (None, self.input_height, self.input_width, 3)
            
        self.target = tf.placeholder(shape=output_shape, dtype=tf.float32)                               
        self.output = self.create_base_model(input=self.input)
        self.output = self.create_deeplab(self.output)
        self.cost = self.soft_dice_loss(self.target, self.output)
        
        if is_training:
            # ---------------------------------- #
            # calculate loss, using soft dice    #
            # ---------------------------------- #
            update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
            with tf.control_dependencies(update_ops):
                 self.optimizer = tf.train.AdamOptimizer(learning_rate=0.0001).minimize(self.cost)
                    
        # ---------------------------------- #
        # tensorflow saver                   #
        # ---------------------------------- #
        self.saver_all = None
        self.saver_partial = None
        self.session = None


        
    def atrous_spatial_pyramid_pooling(self, inputs,
                                        output_stride):
        """Atrous Spatial Pyramid Pooling.
        Args:
            inputs: A tensor of size [batch, height, width, channels].
            output_stride: The ResNet unit's stride. Determines the rates for atrous convolution.
            the rates are (6, 12, 18) when the stride is 16, and doubled when 8.
            batch_norm_decay: The moving average decay when estimating layer activation
            statistics in batch normalization.
            is_training: A boolean denoting whether the input is for training.
            depth: The depth of the ResNet unit output.
        Returns:
            The atrous spatial pyramid pooling output.
        """
        if output_stride not in [8, 16]:
            raise ValueError('output_stride must be either 8 or 16.')
        atrous_rates = [6, 12, 18]

        if output_stride == 8:
            atrous_rates = [2*rate for rate in atrous_rates]

        conv_1x1_1 = self.conv_block(inputs=inputs, 
                                     filters=128, 
                                     kernel_size=(1,1), 
                                     strides=(1,1))

        input_depth = inputs.get_shape().as_list()[-1]
        kernel =  np.random.random((3, 3, input_depth, 128)).astype(np.float32)
        conv_3x3_1 = tf.nn.atrous_conv2d(value=inputs, 
                                         filters=kernel, 
                                         rate=atrous_rates[0], 
                                         padding='SAME', 
                                         name='atrous1')

        conv_3x3_2 = tf.nn.atrous_conv2d(value=inputs, 
                                         filters=kernel, 
                                         rate=atrous_rates[1], 
                                         padding='SAME',
                                         name='atrous2')

        conv_3x3_3 = tf.nn.atrous_conv2d(value=inputs, 
                                        filters=kernel,
                                        rate=atrous_rates[2], 
                                        padding='SAME', 
                                        name='atrous3')
        concat_1 = tf.concat([conv_1x1_1, conv_3x3_1, conv_3x3_2, conv_3x3_3], -1)
        
        conv_1x1_2 = self.conv_block(inputs=concat_1, 
                                     filters=128, 
                                     kernel_size=(1,1), 
                                     strides=(1,1))

        upsample = tf.image.resize_bilinear(images=conv_1x1_2, size=(70,70))
        return upsample
          

    def create_deeplab(self, inputs):
        """[summary]
        
        Arguments:
            inputs {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        with tf.variable_scope("deeplab"):
            inputs = tf.image.resize_bilinear(images=inputs, size=(70,70))
            conv_1x1 = self.conv_block(inputs=inputs, 
                                       filters=128, 
                                       kernel_size=(1,1), 
                                       strides=(1,1), 
                                       padding='same')

            atrous_output = self.atrous_spatial_pyramid_pooling(inputs=inputs, 
                                                                output_stride=16)
            concat_1 = tf.concat([atrous_output, conv_1x1], -1)
            
            if self.num_classes == 1:
                filters = 1
            else:
                filters = 3
            conv_3x3 = self.conv_block(inputs=concat_1, 
                                       filters=filters, 
                                       kernel_size=(1,1), 
                                       strides=(1,1), 
                                       activation=None)
            upsample = tf.image.resize_bilinear(images=conv_3x3, size=(300,300))
            upsample = tf.nn.sigmoid(upsample)
            print (upsample, "==============================>>> final")

        return upsample
    

    def soft_dice_loss(self, y_true, y_pred, epsilon=1e-6):
        """[summary]
        
        Arguments:
            y_true {[type]} -- [description]
            y_pred {[type]} -- [description]
        
        Keyword Arguments:
            epsilon {[type]} -- [description] (default: {1e-6})
        """ 
        numerator = tf.reduce_sum( y_pred * y_true)
        denominator = tf.reduce_sum(tf.square(y_pred)) + tf.reduce_sum(tf.square(y_true)) 
        dice_loss =  1 - (2 * numerator / (denominator + epsilon))

        mse_loss = tf.reduce_mean(tf.square(y_true - y_pred))
        ttl_loss = 0.5 * mse_loss + 0.5 * dice_loss
        return ttl_loss


    def batch_generator(self, batch_size, dataset_path, message='training'):
        """Train Generator
        
        Arguments:
            batch_size {integer} -- the size of the batch
            image_name_list {list of string} -- the list of image name
        """
        label_folder_path = dataset_path + "labels/"
        dataset_folder_path = dataset_path + "images/"
        dataset_file_list = get_filenames(dataset_folder_path)
        random.shuffle(dataset_file_list)
        
        print ("------------------------INFO IMAGES-------------------")
        print ("Image Folder: " + dataset_folder_path)
        print ("Number of Image: " + str(len(dataset_file_list)))
        print ("------------------------------------------------------")

        # Infinite loop.
        idx = 0
        while True:
            x_batch = []
            y_pred = []

            for i in range(batch_size):
                if idx >= len(dataset_file_list):
                    random.shuffle(dataset_file_list)
                    print ("==>>> INFO: your " + message +" dataset is reshuffled again", idx)
                    idx = 0
                
                try:
                    tmp_x = cv2.imread(dataset_folder_path + dataset_file_list[idx])
                    tmp_x = cv2.cvtColor(tmp_x, cv2.COLOR_BGR2RGB)
                    tmp_x = cv2.resize(tmp_x, dsize=(self.input_width, self.input_height), interpolation=cv2.INTER_CUBIC)
                    tmp_x = tmp_x.astype(np.float32) / 255.
                    tmp_y = cv2.imread(label_folder_path + dataset_file_list[idx])
                    tmp_y = cv2.cvtColor(tmp_y, cv2.COLOR_BGR2GRAY)
                    tmp_y = cv2.resize(tmp_y, dsize=(self.input_width, self.input_height), interpolation=cv2.INTER_CUBIC).reshape((300, 300, 1))
                    tmp_y = tmp_y.astype(np.float32) / 255.
                    x_batch.append(tmp_x)
                    y_pred.append(tmp_y)
                except Exception as e:
                    print ("-----------------------------------------------------------------------------")
                    print ('>>> WARNING: fail handling ' +  dataset_file_list[idx], e)
                    print ("-----------------------------------------------------------------------------")

                idx += 1
            yield (np.array(x_batch), np.array(y_pred))


    def optimize(self, subdivisions,
                iterations, 
                best_loss, 
                train_batch, 
                val_batch, 
                save_path):
        """[summary]
        
        Arguments:
            subdivisions {[type]} -- [description]
            iterations {[type]} -- [description]
            best_loss {[type]} -- [description]
            train_batch {[type]} -- [description]
            val_batch {[type]} -- [description]
            save_path {[type]} -- [description]
        """
        
        self.train_losses = []
        self.val_losses = []
        best_loss = best_loss

        for i in range(iterations):
            sign = '-'

            tmp_losses = []
            for j in range(subdivisions):
                # ------------------------- #
                # feed train data           #
                # ------------------------- #
                input_image, target_image = next(train_batch)
                feed_dict = {}
                feed_dict[self.input] = input_image
                feed_dict[self.target] = target_image
                self.session.run(self.optimizer, feed_dict)
                loss = self.session.run(self.cost, feed_dict)
                tmp_losses.append(loss)
                print ("> Train sub", j, 'loss : ', loss)
                
            # ------------------------- #
            # feed validation data      #
            # ------------------------- #
            input_image, target_image = next(val_batch)
            feed_dict = {}
            feed_dict[self.input] = input_image
            feed_dict[self.target] = target_image
            loss = self.session.run(self.cost, feed_dict)

            # ------------------------- #
            # append loss val           #
            # ------------------------- #
            self.val_losses.append(loss)
            train_loss = sum(tmp_losses)/(len(tmp_losses)+0.0001)
            self.train_losses.append(train_loss)
            
            if loss < best_loss:
                best_loss = loss
                sign = '***************'
                self.saver_all.save(self.session, save_path)
                
            print ("> iteration", i, 'train loss: ', train_loss, 'val loss: ', loss, sign)
            
        
        
    def check_val_data(self, val_generator):
        x_val, y_val = next(val_generator)

        val_feed_dict = {}
        val_feed_dict[self.input] = x_val
        val_feed_dict[self.target] = y_val

        loss = self.session.run(self.cost, val_feed_dict)
        print ("> val loss: ", round(loss, 5))
            


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

slim = tf.contrib.slim


class NASBaseCell(object):
  """NASNet Cell class that is used as a 'layer' in image architectures.
  Args:
    num_conv_filters: The number of filters for each convolution operation.
    operations: List of operations that are performed in the NASNet Cell in
      order.
    used_hiddenstates: Binary array that signals if the hiddenstate was used
      within the cell. This is used to determine what outputs of the cell
      should be concatenated together.
    hiddenstate_indices: Determines what hiddenstates should be combined
      together with the specified operations to create the NASNet cell.
  """

  def __init__(self, num_conv_filters, operations, used_hiddenstates,
               hiddenstate_indices, drop_path_keep_prob, total_num_cells,
               total_training_steps):
    assert len(hiddenstate_indices) == len(operations)
    assert len(operations) % 2 == 0
    self._num_conv_filters = num_conv_filters
    self._operations = operations
    self._used_hiddenstates = used_hiddenstates
    self._hiddenstate_indices = hiddenstate_indices
    self._drop_path_keep_prob = drop_path_keep_prob
    self._total_num_cells = total_num_cells
    self._total_training_steps = total_training_steps

  def __call__(self, net, scope, filter_scaling, stride, prev_layer, cell_num):
    self._cell_num = cell_num
    self._filter_scaling = filter_scaling
    self._filter_size = int(self._num_conv_filters * filter_scaling)

    with tf.variable_scope(scope):
      net = self._cell_base(net, prev_layer)
      for i in range(int(len(self._operations) / 2)):
        with tf.variable_scope('comb_iter_{}'.format(i)):
          h1 = net[self._hiddenstate_indices[i * 2]]
          h2 = net[self._hiddenstate_indices[i * 2 + 1]]
          with tf.variable_scope('left'):
            h1 = self._apply_operation(h1, self._operations[i * 2], stride, 
                                       self._hiddenstate_indices[i * 2] < 2)
          with tf.variable_scope('right'):
            h2 = self._apply_operation(h2, self._operations[i * 2 + 1], stride,
                                       self._hiddenstate_indices[i * 2 + 1] < 2)
          with tf.variable_scope('combine'):
            h = h1 + h2
          net.append(h)

      with tf.variable_scope('cell_output'):
        net = self._combine_unused_states(net)

      return net

  def _cell_base(self, net, prev_layer):
    filter_size = self._filter_size

    if prev_layer is None:
      prev_layer = net
    elif net.shape[2] != prev_layer.shape[2]:
      prev_layer = tf.nn.relu(prev_layer)
      prev_layer = self._factorized_reduction(prev_layer, filter_size, stride=2)
    elif filter_size != prev_layer.shape[3]:
      prev_layer = tf.nn.relu(prev_layer)
      prev_layer = slim.conv2d(prev_layer, filter_size, 1, scope='prev_1x1')
      prev_layer = slim.batch_norm(prev_layer, scope='prev_bn')

    net = tf.nn.relu(net)
    net = slim.conv2d(net, filter_size, 1, scope='1x1')
    net = slim.batch_norm(net, scope='beginning_bn')
    net = tf.split(axis=3, num_or_size_splits=1, value=net)
    for split in net:
      assert split.shape[3] == filter_size
    net.append(prev_layer)
    return net

  def _apply_operation(self, net, operation, stride, is_from_original_input):
    if stride > 1 and not is_from_original_input:
      stride = 1
    input_filters = net.shape[3]
    filter_size = self._filter_size
    if 'separable' in operation:
      num_layers = int(operation.split('_')[-1])
      kernel_size = int(operation.split('x')[0][-1])
      for layer_num in range(num_layers):
        net = tf.nn.relu(net)
        net = slim.separable_conv2d(
            net,
            filter_size,
            kernel_size,
            depth_multiplier=1,
            scope='separable_{0}x{0}_{1}'.format(kernel_size, layer_num + 1),
            stride=stride)
        net = slim.batch_norm(
            net, scope='bn_sep_{0}x{0}_{1}'.format(kernel_size, layer_num + 1))
        stride = 1
    elif operation in ['none']:
      if stride > 1 or (input_filters != filter_size):
        net = tf.nn.relu(net)
        net = slim.conv2d(net, filter_size, 1, stride=stride, scope='1x1')
        net = slim.batch_norm(net, scope='bn_1')
    elif 'pool' in operation:
      pooling_type = operation.split('_')[0]
      pooling_shape = int(operation.split('_')[-1].split('x')[0])
      if pooling_type == 'avg':
        net = slim.avg_pool2d(net, pooling_shape, stride=stride, padding='SAME')
      elif pooling_type == 'max':
        net = slim.max_pool2d(net, pooling_shape, stride=stride, padding='SAME')
      else:
        raise ValueError('Unimplemented pooling type: ', pooling_type)
      if input_filters != filter_size:
        net = slim.conv2d(net, filter_size, 1, stride=1, scope='1x1')
        net = slim.batch_norm(net, scope='bn_1')
    else:
      raise ValueError('Unimplemented operation: ', operation)

    #if operation != 'none':
      #net = self._apply_drop_path(net)
    return net

  def _combine_unused_states(self, net):
    used_hiddenstates = self._used_hiddenstates
    states_to_combine = (
        [h for h, is_used in zip(net, used_hiddenstates) if not is_used])
    net = tf.concat(values=states_to_combine, axis=3)
    return net

  def _apply_drop_path(self, net):
    drop_path_keep_prob = self._drop_path_keep_prob
    if drop_path_keep_prob < 1.0:
      # Scale keep prob by layer number
      assert self._cell_num != -1
      layer_ratio = (self._cell_num + 1) / float(self._total_num_cells)
      drop_path_keep_prob = 1 - layer_ratio * (1 - drop_path_keep_prob)
      # Decrease keep prob over time
      current_step = tf.cast(tf.train.get_or_create_global_step(), tf.float32)
      current_ratio = tf.minimum(1.0, current_step / self._total_training_steps)
      drop_path_keep_prob = 1 - current_ratio * (1 - drop_path_keep_prob)
      # Drop path
      noise_shape = [None, 1, 1, 1]
      random_tensor = drop_path_keep_prob
      print ('..........', noise_shape)
      random_tensor += tf.random_uniform(noise_shape, dtype=tf.float32)
      binary_tensor = tf.cast(tf.floor(random_tensor), net.dtype)
      keep_prob_inv = tf.cast(1.0 / drop_path_keep_prob, net.dtype)
      net = net * keep_prob_inv * binary_tensor
    return net

  def _factorized_reduction(self, net, output_filters, stride):
    assert output_filters % 2 == 0
    if stride == 1:
      net = slim.conv2d(net, output_filters, 1, scope='path_conv')
      net = slim.batch_norm(net, scope='path_bn')
      return net
    stride_spec = [1, stride, stride, 1]

    # Skip path 1
    path1 = tf.nn.avg_pool(net, [1, 1, 1, 1], stride_spec, 'VALID')
    path1 = slim.conv2d(path1, int(output_filters / 2), 1, scope='path1_conv')

    # Skip path 2
    pad_arr = [[0, 0], [0, 1], [0, 1], [0, 0]]
    path2 = tf.pad(net, pad_arr)[:, 1:, 1:, :]
    path2 = tf.nn.avg_pool(path2, [1, 1, 1, 1], stride_spec, 'VALID')
    path2 = slim.conv2d(path2, int(output_filters / 2), 1, scope='path2_conv')

    # Concat and apply BN
    final_path = tf.concat(values=[path1, path2], axis=3)
    final_path = slim.batch_norm(final_path, scope='final_path_bn')
    return final_path


class PNASCell(NASBaseCell):
  """PNASNet Cell."""

  def __init__(self, num_conv_filters, drop_path_keep_prob, total_num_cells,
               total_training_steps):
    # Configuration for the PNASNet-5 model.
    operations = [
        'separable_5x5_2', 'max_pool_3x3', 'separable_7x7_2', 'max_pool_3x3',
        'separable_5x5_2', 'separable_3x3_2', 'separable_3x3_2', 'max_pool_3x3',
        'separable_3x3_2', 'none'
    ]
    used_hiddenstates = [1, 1, 0, 0, 0, 0, 0]
    hiddenstate_indices = [1, 1, 0, 0, 0, 0, 4, 0, 1, 0]

    super(PNASCell, self).__init__(
        num_conv_filters, operations, used_hiddenstates, hiddenstate_indices,
        drop_path_keep_prob, total_num_cells, total_training_steps)



import json
import cv2
import math
import random
import numpy as np
import tensorflow as tf
import qoalai.networks.modified_densenet as densenet
from simple_tensor.tensor_operations import *
from simple_tensor.tensor_losses import mse_loss_mean, mse_loss_sum
from comdutils.file_utils import *
slim = tf.contrib.slim



class Landmark(object):
    def __init__(self,
                 num_landmark_point,
                 input_height = 300,
                 input_width = 300, 
                 input_channel = 3,
                 grid_num = 19,
                 threshold = 0.7,
                 pos_loss_alpha = 1.,
                 neg_loss_alpha = 1.2,
                 position_loss_alpha = 1.):

        """Constructor
        
        Arguments:
            classes {list of string} -- the image classes list

        Keyword Arguments:
            input_height {int} -- the height of input image (default: {512})
            input_width {int} -- the width of input image (default: {512})
            input_channel {int} -- the channel of input image (default: {3})
        """
        self.num_landmark_point = num_landmark_point
        self.input_height = input_height
        self.input_width = input_width
        self.input_channel = input_channel
        self.grid_num = grid_num
        self.grid_size_relative = 1./self.grid_num
        self.threshold = threshold

        self.pos_loss_alpha = pos_loss_alpha
        self.neg_loss_alpha = neg_loss_alpha
        self.position_loss_alpha = position_loss_alpha

        self.input_placeholder = tf.placeholder(tf.float32, shape=(None, self.input_height, self.input_width, self.input_channel))
        self.point_position_placeholder = tf.placeholder(tf.float32, shape=(None, self.grid_num, self.grid_num, 2), name='position_ph')
        self.point_map_placeholder = tf.placeholder(tf.float32, shape=(None, self.grid_num, self.grid_num, 1), name='map_ph')

        self.grid_position_mask_onx_np = np.zeros((1, self.grid_num , self.grid_num, 1))
        self.grid_position_mask_ony_np = np.zeros((1, self.grid_num , self.grid_num, 1))

        for j in range(self.grid_num):
            for k in range(self.grid_num):
                self.grid_position_mask_onx_np[:, j, k, :] = k
                self.grid_position_mask_ony_np[:, j, k, :] = j
        
        self.grid_position_mask_onx = tf.convert_to_tensor(self.grid_position_mask_onx_np, dtype=tf.float32)
        self.grid_position_mask_ony = tf.convert_to_tensor(self.grid_position_mask_ony_np, dtype=tf.float32)

    
    def build_densenet_base(self, input_tensor,
                            dropout_rate,
                            is_training,
                            top_layer_depth):
        """[summary]
        
        Arguments:
            input_tensordropout_rate {[type]} -- [description]
            is_training {bool} -- [description]
        """
        # ----------------------------------- #
        # building the densenet base network  #
        # ----------------------------------- #
        arg_scoope = densenet.densenet_arg_scope()
        out = None

        with slim.arg_scope(arg_scoope):
            out = densenet.densenet121(inputs=input_tensor, 
                                       num_classes=1, 
                                       is_training=is_training)
            base_var_list = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)
        
        print ('===========', out)
        # 12x12x3
        out = new_conv2d_layer(input=out, 
                               filter_shape=[2, 2, out.get_shape().as_list()[-1], 3], 
                               dropout_val=1.-dropout_rate, 
                               activation = 'NONE', 
                               padding='SAME', 
                               strides=[1, 1, 1, 1],
                               is_training=is_training,
                               use_bias=False,
                               use_batchnorm=False) 
        print ("===>>>", out)
        self.point_map = tf.nn.sigmoid(out[:, :, :, :1])
        self.point_position = tf.nn.tanh(out[:, :, :, 1:])
        
        print ('--------------------INFO---------------------')
        print ('The shape of your point map: ', self.point_map)
        print ('The shape of your point position: ', self.point_position)
        print ('---------------------------------------------')


    def calculate_loss(self):
        """[summary]
        
        Returns:
            [type] -- [description]
        """
        # ----------------
        # the output map ideally describe the error probaility between the predicted point with the label point
        # the output map placeholder is the distance for each grid top left corner to the target points
        # the label for output map is the the distance between the predicted point with the label point
        # to make it as a probability, we convert the label for output map to : 1. - relative distance to image size
        # ---------------
        pred_mapx = self.point_position[:, :, :, :1] * self.grid_size_relative 
        pred_mapy = self.point_position[:, :, :, 1:] * self.grid_size_relative 

        label_mapx = self.point_position_placeholder[:, :, :, :1] * self.grid_size_relative
        label_mapy = self.point_position_placeholder[:, :, :, 1:] * self.grid_size_relative
        label_deltax = label_mapx - pred_mapx
        label_deltay = label_mapy - pred_mapy
        label_map = 1. - tf.sqrt(tf.square(label_deltax) + tf.square(label_deltay))

        # ----------------
        # positive map error
        # ----------------
        label_map_pos_tmp = tf.math.greater(label_map, tf.convert_to_tensor(np.array(self.threshold), tf.float32))
        label_map_pos_tmp = tf.cast(label_map_pos_tmp, tf.float32)

        label_map_pos = label_map * label_map_pos_tmp 
        pred_map_pos = self.point_map * label_map_pos_tmp
        self.point_map_loss_pos = self.pos_loss_alpha * mse_loss_mean(pred_map_pos, label_map_pos)

        # ----------------
        # negative map error
        # ----------------
        label_map_neg_tmp = 1. - label_map_pos_tmp

        label_map_neg = label_map * label_map_neg_tmp 
        pred_map_neg = self.point_map * label_map_neg_tmp
        self.point_map_loss_neg = self.neg_loss_alpha * mse_loss_mean(pred_map_neg, label_map_neg)
        
        # ----------------
        # position error
        # ----------------
        tmp_point_map = tf.math.greater(self.point_map_placeholder, tf.convert_to_tensor(np.array(self.threshold), tf.float32))
        tmp_point_map = tf.cast(tmp_point_map, tf.float32)

        tmp_point_position = self.point_position * tmp_point_map
        tmp_point_position_placeholder = self.point_position_placeholder * tmp_point_map
        self.point_position_loss = self.position_loss_alpha *  mse_loss_sum(tmp_point_position, tmp_point_position_placeholder) / tf.math.reduce_sum(tmp_point_map)

        # ---------------
        # total loss
        # ---------------
        self.total_loss = self.point_map_loss_pos + self.point_map_loss_neg + self.point_position_loss
        
        tmp_map_label = tf.math.greater(self.point_map_placeholder, tf.convert_to_tensor(np.array(self.threshold), tf.float32))
        tmp_map_label = tf.cast(tmp_map_label, tf.float32)
        self.point_map_acc = self.object_accuracy(self.point_map, tmp_map_label)
        return self.total_loss


    def object_accuracy(self, objectness_pred, objectness_label):
        """[summary]
        
        Arguments:
            objectness_pred {[type]} -- [description]
            objectness_label {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        objectness_mask = tf.math.greater(objectness_pred, tf.convert_to_tensor(np.array(self.threshold), tf.float32))
        objectness_mask = tf.cast(objectness_mask, tf.float32)
        delta = (objectness_label - objectness_mask) * objectness_label
        obj_acc = 1. - tf.reduce_sum(delta) / (tf.reduce_sum(objectness_label) + 0.0001)
        return obj_acc


    def read_target(self, file_path):
        # ----------------------------------- #
        # create initial label for point map and position   
        # ----------------------------------- #
        point_map_lable = np.zeros((self.grid_num, self.grid_num, 3), dtype=np.float32)

        # ----------------------------------- #
        # reading the label in a .txt file foreach image 
        #------------------------------------ #
        x_points = []
        y_points = []
        
        file = open(file_path, "r") 
        data = file.read()
        data = data.split()
        length = len(data)

        if length != 2 * self.num_landmark_point:
            print ("-------------------WARNING-------------------")
            print ("WRONG label format")
            print ("the label format is x1, y1, x2, y2, ..., xn, yn")
            print ("the value of x and y are relative to image width and height")
            print ("each image has single .txt file as label with the name")
            print (file_path)
            print ("---------------------------------------------")

        for i in range(self.num_landmark_point):
            data_x = float(data[i*2])
            data_y = float(data[i*2 + 1])
            x_points.append(data_x)
            y_points.append(data_y)
            
        for i in range(self.grid_num):
            for j in range(self.grid_num):
                biggest_val = 0.
                nearest_point = 0

                for k in range(self.num_landmark_point):
                    # ----------------------------------- #
                    # filling the point map label         #
                    #------------------------------------ #
                    tmpx = abs(x_points[k] - (j + 0.5) * self.grid_size_relative)
                    tmpy = abs(y_points[k] - (i + 0.5) * self.grid_size_relative)
                    tmp = 1. - math.sqrt(tmpx * tmpx + tmpy * tmpy)

                    if tmp > biggest_val:
                        biggest_val = tmp
                        nearest_point = k 
                
                x = x_points[nearest_point]
                y = y_points[nearest_point]
                point_map_lable[i, j, 0] = biggest_val
                point_map_lable[i, j, 1] = x /self.grid_size_relative - j 
                point_map_lable[i, j, 2] = y /self.grid_size_relative - i 
                
        return point_map_lable[:, :, :1], point_map_lable[:, :, 1:]


    def batch_generator(self, batch_size, dataset_path, message):
        """Train Generator
        
        Arguments:
            batch_size {integer} -- the size of the batch
            image_name_list {list of string} -- the list of image name
        """
        label_folder_path = dataset_path + "labels/"
        dataset_folder_path = dataset_path + "images/"
        dataset_file_list = get_filenames(dataset_folder_path)
        random.shuffle(dataset_file_list)
        
        print ("------------------------INFO-------------------")
        print ("Image Folder: " + dataset_folder_path)
        print ("Number of Image: " + str(len(dataset_file_list)))
        print ("-----------------------------------------------")

        # Infinite loop.
        idx = 0
        while True:
            x_batch = []
            y_pred_map = []
            y_pred_position = []

            for i in range(batch_size):
                if idx >= len(dataset_file_list):
                    random.shuffle(dataset_file_list)
                    print ("==>>> INFO: your " + message +" dataset is reshuffled again: ", idx)
                    idx = 0
                    
                try:
                    tmp_x = cv2.imread(dataset_folder_path + dataset_file_list[idx])
                    tmp_x = cv2.cvtColor(tmp_x, cv2.COLOR_BGR2RGB)
                    tmp_x = cv2.resize(tmp_x, (self.input_width, self.input_height))
                    tmp_x = tmp_x.astype(np.float32) / 255.
                    tmp_y_map, tmp_y_position = self.read_target(label_folder_path + dataset_file_list[idx].split('.')[0] + ".txt")
                    x_batch.append(tmp_x)
                    y_pred_map.append(tmp_y_map)
                    y_pred_position.append(tmp_y_position)

                except Exception as e:
                    print ("---------------------------------------------------------------")
                    print ("WARNING: the " + str(e))
                    print ("---------------------------------------------------------------")

                idx += 1
            yield (np.array(x_batch), np.array(y_pred_map), np.array(y_pred_position))


    def optimize(self, 
                 iteration, 
                 subdivition,
                 cost_tensor,
                 optimizer_tensor,
                 session,
                 saver, 
                 train_generator,
                 val_generator,
                 best_loss,
                 path_tosave_model='model/model1'):
        """[summary]
        
        Arguments:
            iteration {[type]} -- [description]
            subdivition {[type]} -- [description]
            cost_tensor {[type]} -- [description]
            optimizer_tensor {[type]} -- [description]
            out_tensor {[type]} -- [description]
        
        Keyword Arguments:
            train_batch_size {int} -- [description] (default: {32})
            val_batch_size {int} -- [description] (default: {50})
            path_tosave_model {str} -- [description] (default: {'model/model1'})
        """

        self.train_loss = []
        self.val_loss = []
        best_loss = best_loss
        
        for i in range(iteration):
            sign = "-"
            t_losses = []

            for j in range(subdivition):
                x_train, y_train_map, y_train_position = next(train_generator)
                feed_dict = {}
                feed_dict[self.input_placeholder] = x_train
                feed_dict[self.point_map_placeholder] = y_train_map
                feed_dict[self.point_position_placeholder] = y_train_position
                session.run(optimizer_tensor, feed_dict)
                loss = session.run(cost_tensor, feed_dict)
                t_losses.append(loss)
                print ("> Train sub", j, 'loss : ', loss)
                
            x_val, y_val_map, y_val_position = next(val_generator)
            feed_dict = {}
            feed_dict[self.input_placeholder] = x_val
            feed_dict[self.point_map_placeholder] = y_val_map
            feed_dict[self.point_position_placeholder] = y_val_position
            loss = session.run(cost_tensor, feed_dict)
            map_acc, pml_pos, pml_neg, ppl = session.run([self.point_map_acc, self.point_map_loss_pos, self.point_map_loss_neg, self.point_position_loss], feed_dict)
            
            t_loss = sum(t_losses) / (len(t_losses) + 0.0001)
            self.train_loss.append(t_loss)
            self.val_loss.append(loss)
                
            if best_loss > loss:
                best_loss = loss
                sign = "************* model saved"
                saver.save(session, path_tosave_model)
        
            print (">> epoch: ", i, "train loss: ", round(t_loss, 3), "val loss: ", round(loss, 3), \
                "val map acc: ", round(map_acc, 3), "val map pos loss: ", round(pml_pos, 3), "val map neg loss: ", round(pml_neg, 3), "val poss loss: ", round(ppl, 3), sign)  



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the definition for the PNASNet classification networks.
Paper: https://arxiv.org/abs/1712.00559
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
from qoalai.networks.cell import PNASCell

arg_scope = tf.contrib.framework.arg_scope
slim = tf.contrib.slim


# Notes for training large PNASNet model on ImageNet
# -------------------------------------
# batch size (per replica): 16
# learning rate: 0.015 * 100
# learning rate decay factor: 0.97
# num epochs per decay: 2.4
# sync sgd with 100 replicas
# auxiliary head loss weighting: 0.4
# label smoothing: 0.1
# clip global norm of all gradients by 10
def large_imagenet_config():
  """Large ImageNet configuration based on PNASNet-5."""
  return tf.contrib.training.HParams(
      stem_multiplier=3.0,
      dense_dropout_keep_prob=0.5,
      num_cells=12,
      filter_scaling_rate=2.0,
      num_conv_filters=216,
      drop_path_keep_prob=0.6,
      use_aux_head=0, #1
      num_reduction_layers=2,
      total_training_steps=250000,
      num_stem_cells=2,
  )


def calc_reduction_layers(num_cells, num_reduction_layers):
  """Figure out what layers should have reductions."""
  reduction_layers = []
  for pool_num in range(1, num_reduction_layers + 1):
    layer_num = (float(pool_num) / (num_reduction_layers + 1)) * num_cells
    layer_num = int(layer_num)
    reduction_layers.append(layer_num)
  return reduction_layers


def pnasnet_large_arg_scope(weight_decay=4e-5,
                            batch_norm_decay=0.9997,
                            batch_norm_epsilon=1e-3):
  batch_norm_params = {
      'decay': batch_norm_decay, # decay for the moving averages
      'epsilon': batch_norm_epsilon, # epsilon to prevent 0s in variance
      'scale': True,
      'fused': True,
  }
  weights_regularizer = tf.contrib.layers.l2_regularizer(weight_decay)
  weights_initializer = tf.contrib.layers.variance_scaling_initializer(
      mode='FAN_OUT')
  with arg_scope([slim.fully_connected, slim.conv2d, slim.separable_conv2d],
                 weights_regularizer=weights_regularizer,
                 weights_initializer=weights_initializer):
    with arg_scope([slim.fully_connected],
                   activation_fn=None, scope='FC'):
      with arg_scope([slim.conv2d, slim.separable_conv2d],
                     activation_fn=None, biases_initializer=None):
        with arg_scope([slim.batch_norm], **batch_norm_params) as sc:
          return sc


def _imagenet_stem(inputs, hparams, stem_cell):
  """Stem used for models trained on ImageNet."""
  num_stem_filters = int(32 * hparams.stem_multiplier)
  net = slim.conv2d(inputs, num_stem_filters, [3, 3], stride=2, scope='conv0',
      padding='VALID')
  net = slim.batch_norm(net, scope='conv0_bn')

  # Run the reduction cells
  cell_outputs = [None, net]
  filter_scaling = 1.0 / (hparams.filter_scaling_rate**hparams.num_stem_cells)
  for cell_num in range(hparams.num_stem_cells):
    net = stem_cell(
        net,
        scope='cell_stem_{}'.format(cell_num),
        filter_scaling=filter_scaling,
        stride=2,
        prev_layer=cell_outputs[-2],
        cell_num=cell_num)
    cell_outputs.append(net)
    filter_scaling *= hparams.filter_scaling_rate
  return net, cell_outputs


def _build_aux_head(net, end_points, num_classes, hparams, scope):
  """Auxiliary head used for all models across all datasets."""
  with tf.variable_scope(scope):
    aux_logits = tf.identity(net)
    with tf.variable_scope('aux_logits'):
      aux_logits = slim.avg_pool2d(
          aux_logits, [5, 5], stride=3, padding='VALID')
      aux_logits = slim.conv2d(aux_logits, 128, [1, 1], scope='proj')
      aux_logits = slim.batch_norm(aux_logits, scope='aux_bn0')
      aux_logits = tf.nn.relu(aux_logits)
      # Shape of feature map before the final layer.
      shape = aux_logits.shape
      shape = shape[1:3]
      aux_logits = slim.conv2d(aux_logits, 768, shape, padding='VALID')
      aux_logits = slim.batch_norm(aux_logits, scope='aux_bn1')
      aux_logits = tf.nn.relu(aux_logits)
      aux_logits = tf.contrib.layers.flatten(aux_logits)
      aux_logits = slim.fully_connected(aux_logits, num_classes)
      end_points['AuxLogits'] = aux_logits


def build_pnasnet_large(images,
                        num_classes,
                        is_training=True,
                        final_endpoint=None,
                        config=None):
  """Build PNASNet Large model for the ImageNet Dataset."""
  hparams = large_imagenet_config()
  if not is_training:
    hparams.set_hparam('drop_path_keep_prob', 1.0)

  total_num_cells = hparams.num_cells + hparams.num_stem_cells
  cell = PNASCell(hparams.num_conv_filters, hparams.drop_path_keep_prob,
                  total_num_cells, hparams.total_training_steps)

  with arg_scope([slim.dropout, slim.batch_norm], is_training=is_training):
    end_points = {}
    def add_and_check_endpoint(endpoint_name, net):
      end_points[endpoint_name] = net
      return final_endpoint and (endpoint_name == final_endpoint)

    # Find where to place the reduction cells or stride normal cells
    reduction_indices = calc_reduction_layers(
        hparams.num_cells, hparams.num_reduction_layers)

    net, cell_outputs = _imagenet_stem(images, hparams, cell)
    if add_and_check_endpoint('Stem', net):
      return net, end_points

    # Setup for building in the auxiliary head.
    aux_head_cell_idxes = []
    if len(reduction_indices) >= 2:
      aux_head_cell_idxes.append(reduction_indices[1] - 1)

    # Run the cells
    filter_scaling = 1.0
    for cell_num in range(hparams.num_cells):
      is_reduction = cell_num in reduction_indices
      stride = 2 if is_reduction else 1
      if is_reduction: filter_scaling *= hparams.filter_scaling_rate
      net = cell(
          net,
          scope='cell_{}'.format(cell_num),
          filter_scaling=filter_scaling,
          stride=stride,
          prev_layer=cell_outputs[-2],
          cell_num=hparams.num_stem_cells + cell_num)
      if add_and_check_endpoint('Cell_{}'.format(cell_num), net):
        return net, end_points
      cell_outputs.append(net)

      if (hparams.use_aux_head and cell_num in aux_head_cell_idxes and
          num_classes and is_training):
        aux_net = tf.nn.relu(net)
        _build_aux_head(aux_net, end_points, num_classes, hparams,
                        scope='aux_{}'.format(cell_num))

    # Final softmax layer
    with tf.variable_scope('final_layer'):
      net = tf.nn.relu(net)
      net = tf.reduce_mean(net, [1, 2])
      if add_and_check_endpoint('global_pool', net) or not num_classes:
        return net, end_points

      net = slim.dropout(net, hparams.dense_dropout_keep_prob, scope='dropout')
      logits = slim.fully_connected(net, num_classes)
      if add_and_check_endpoint('Logits', logits):
        return net, end_points

      predictions = tf.nn.softmax(logits, name='predictions')
      if add_and_check_endpoint('Predictions', predictions):
        return net, end_points

    return logits, end_points



'''
    File name: detector_utils.py
    Author: [Mochammad F Rahman]
    Date created: / /2019
    Date last modified:  / /2019
    Python Version: >= 3.5
    qoalai version: v0.4
    License: MIT License
    Maintainer: [Mochammad F Rahman]
'''

import json
import cv2
import random
import numpy as np
import tensorflow as tf
from qoalai.tensor_operations import *
from qoalai.networks.inception_utils import *
from qoalai.networks.inception_v4 import *
from qoalai.networks.resnet_v2 import *
from qoalai.networks.vgg import *
from qoalai.networks.nasnet5 import * 
import qoalai.networks.densenet as densenet
from comdutils.file_utils import *


class ImageRecognition(object):
    def __init__(self,
                 classes, 
                 input_height = 400,
                 input_width = 400, 
                 input_channel = 3):

        """Constructor
        
        Arguments:
            classes {list of string} -- the image classes list

        Keyword Arguments:
            input_height {int} -- the height of input image (default: {512})
            input_width {int} -- the width of input image (default: {512})
            input_channel {int} -- the channel of input image (default: {3})
        """

        self.classes = classes
        self.input_height = input_height
        self.input_width = input_width
        self.input_channel = input_channel

        self.input_placeholder = tf.placeholder(tf.float32, shape=(None, self.input_height, self.input_width, self.input_channel))
        self.output_placeholder = tf.placeholder(tf.float32, shape=(None, len(self.classes)))


    def build_resnetv2(self, 
                       input_tensor,
                       is_training,
                       top_layer_depth = 128): 
        """[summary]
        
        Arguments:
            input_tensor {[type]} -- [description]
            is_training {bool} -- [description]
        
        Keyword Arguments:
            top_layer_depth {int} -- [description] (default: {128})
        
        Returns:
            [type] -- [description]
        """
        
        out = None
        with slim.arg_scope(resnet_arg_scope()):
            out, end_points = resnet_v2_101(inputs = input_tensor,
                                            num_classes=1001,
                                            is_training=is_training,
                                            global_pool=True,
                                            output_stride=None,
                                            spatial_squeeze=True,
                                            reuse=None,
                                            scope='resnet_v2_101')
            base_var_list = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)
        
        with tf.variable_scope('resnet_v2_101'):
            depth = out.get_shape().as_list()[-1]
            out = new_fc_layer(out, 
                                num_inputs = depth, 
                                num_outputs = len(self.classes), 
                                name = 'fc1', 
                                dropout_val=1, 
                                activation="NONE",
                                lrelu_alpha=0.2, 
                                data_type=tf.float32,
                                is_training=is_training,
                                use_bias=False)

            if len(self.classes) == 1:
                out = tf.nn.sigmoid(out)
            else:
                out = tf.nn.softmax(out)

        return out, base_var_list


    def build_inceptionv4_basenet(self, 
                                  input_tensor, 
                                  is_training = False, 
                                  final_endpoint='Mixed_7d',
                                  top_layer_depth = 128):
        """Fucntion for creating inception v4 base network
        
        Arguments:
            input_tensor {tensorflow tensor} -- The input tensor
            is_training {bool} -- training or not 
        
        Returns:
            [type] -- [description]
        """
        print ('-------------------------------------------------------')
        print (" NOTICE, your inception v4 base model is end with node:")
        print (final_endpoint)
        print ('-------------------------------------------------------')

        inception_v4_arg_scope = inception_arg_scope
        arg_scope = inception_v4_arg_scope()
        # build inception v4 base graph
        with slim.arg_scope(arg_scope):
            # get output (logits)
            out, end_points = inception_v4(input_tensor, 
                                           num_classes=1, 
                                           final_endpoint=final_endpoint, 
                                           is_training=is_training)
            # get inception variable name
            base_var_list = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)

        with tf.variable_scope('InceptionV4'):
            size = out.get_shape().as_list()[1]
            while(True):
                if size == 1:
                    break

                out = new_conv2d_layer(out, 
                                    filter_shape=[3, 3, out.get_shape().as_list()[-1], top_layer_depth], 
                                    name='cv1', 
                                    dropout_val=0.85, 
                                    activation = 'LRELU', 
                                    lrelu_alpha=0.2,
                                    padding='SAME', 
                                    strides=[1, 2, 2, 1],
                                    data_type=tf.float32,  
                                    is_training=is_training,
                                    use_bias=True,
                                    use_batchnorm=True) 
                size = out.get_shape().as_list()[1]
            
            depth = out.get_shape().as_list()[-1]
            out = tf.reshape(out, [tf.shape(out)[0], -1])
            out = new_fc_layer(out, 
                                num_inputs = depth, 
                                num_outputs = len(self.classes), 
                                name = 'fc1', 
                                dropout_val=1, 
                                activation="NONE",
                                lrelu_alpha=0.2, 
                                data_type=tf.float32,
                                is_training=is_training,
                                use_bias=False)

            if len(self.classes) == 1:
                out = tf.nn.sigmoid(out)
            else:
                out = tf.nn.softmax(out)
            
            self.out = out
        return out, base_var_list


    def build_densenet_base(self, input_tensor,
                            dropout_rate,
                            is_training,
                            top_layer_depth = 128):
        """[summary]
        
        Arguments:
            input_tensordropout_rate {[type]} -- [description]
            is_training {bool} -- [description]
        """
        arg_scoope = densenet.densenet_arg_scope()
        with slim.arg_scope(arg_scoope):
            out = densenet.densenet121(inputs=input_tensor, 
                                       num_classes=1, 
                                       is_training=is_training)
            base_var_list = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)

        with tf.variable_scope('densenet121'):
            size = out.get_shape().as_list()[1]
            while(True):
                if size == 1:
                    break

                out = new_conv2d_layer(out, 
                                        filter_shape=[3, 3, out.get_shape().as_list()[-1], top_layer_depth], 
                                        name='cv1', 
                                        dropout_val=0.85, 
                                        activation = 'LRELU', 
                                        lrelu_alpha=0.2,
                                        padding='SAME', 
                                        strides=[1, 2, 2, 1],
                                        data_type=tf.float32,  
                                        is_training=is_training,
                                        use_bias=True,
                                        use_batchnorm=True) 
                size = out.get_shape().as_list()[1]
            
            depth = out.get_shape().as_list()[-1]
            out = tf.reshape(out, [tf.shape(out)[0], -1])
            out = new_fc_layer(out, 
                                num_inputs = depth, 
                                num_outputs = len(self.classes), 
                                name = 'fc1', 
                                dropout_val=1, 
                                activation="NONE",
                                lrelu_alpha=0.2, 
                                data_type=tf.float32,
                                is_training=is_training,
                                use_bias=False)

            if len(self.classes) == 1:
                out = tf.nn.sigmoid(out)
            else:
                out = tf.nn.softmax(out)

            self.out = out
        return out, base_var_list 


    def build_pnasnet_large_base(self, input_tensor,
                                is_training): 

        out, _ = build_pnasnet_large(images=input_tensor,
                                    num_classes = len(self.classes),
                                    is_training=is_training,
                                    final_endpoint=None,
                                    config=None)
        base_var_list = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES) 
        
        print ("============", out)
        with tf.variable_scope('pnasnet_large'):
            depth = out.get_shape().as_list()[-1]
            out = new_fc_layer(out, 
                                num_inputs = depth, 
                                num_outputs = len(self.classes), 
                                name = 'fc1', 
                                dropout_val=1, 
                                activation="NONE",
                                lrelu_alpha=0.2, 
                                data_type=tf.float32,
                                is_training=is_training,
                                use_bias=False)
            print ('..............', out)
            if len(self.classes) == 1:
                out = tf.nn.sigmoid(out)
            else:
                out = tf.nn.softmax(out)
        return out, base_var_list


    def build_vgg19_base(self, input_tensor,
                            dropout_rate,
                            is_training):
        """[summary]
        
        Arguments:
            input_tensordropout_rate {[type]} -- [description]
            is_training {bool} -- [description]
        """
        arg_scope = vgg_arg_scope()
        with slim.arg_scope(arg_scope):
            out, _ = vgg_19(inputs=input_tensor, 
                        dropout_keep_prob=1. - dropout_rate, 
                        is_training=is_training,
                        spatial_squeeze=False,
                        global_pool=True)
            base_var_list = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)

        with tf.variable_scope('vgg_19'):
            print ('..............', out)
            depth = out.get_shape().as_list()[-1] * out.get_shape().as_list()[-2] * out.get_shape().as_list()[-3]
            out = tf.reshape(out, [tf.shape(out)[0], -1])
            print ('..............', out, depth)
            out = new_fc_layer(out, 
                                num_inputs = depth, 
                                num_outputs = len(self.classes), 
                                name = 'fc1', 
                                dropout_val=1, 
                                activation="NONE",
                                lrelu_alpha=0.2, 
                                data_type=tf.float32,
                                is_training=is_training,
                                use_bias=False)
            print ('..............', out)
            if len(self.classes) == 1:
                out = tf.nn.sigmoid(out)
            else:
                out = tf.nn.softmax(out)

        return out, base_var_list
        

    def batch_generator(self, batch_size, dataset_path, message):
        """Train Generator
        
        Arguments:
            batch_size {integer} -- the size of the batch
            image_name_list {list of string} -- the list of image name
        """
        file_list_by_class = {}
        idx = {}
        for i in self.classes:
            file_list_by_class[i] = get_filenames(dataset_path + i)
            random.shuffle(file_list_by_class[i])
            idx[i] = 0
        
        perclass_sample = int(batch_size/len(self.classes))

        print ("------------------------INFO IMAGES-------------------")
        print ("Image Folder: " + dataset_path)
        for i in self.classes:
            print ("Number of Image in " + str(i) + ": ", len(file_list_by_class[i]))
        print ("------------------------------------------------------")

        # Infinite loop.
        while True:
            x_batch = []
            y_pred = []

            for i in self.classes:
                for j in range(perclass_sample):
                    if idx[i] >= len(file_list_by_class[i]):
                        random.shuffle(file_list_by_class[i])
                        print ("==>>> INFO: your " + message + " in class " + str(i) + " dataset is reshuffled again", idx[i])
                        idx[i] = 0
                    try:
                        tmp_x = cv2.imread(dataset_path + i + "/" + file_list_by_class[i][idx[i]])
                        tmp_x = cv2.cvtColor(tmp_x, cv2.COLOR_BGR2RGB)
                        tmp_x = cv2.resize(tmp_x, dsize=(self.input_width, self.input_height), interpolation=cv2.INTER_CUBIC)
                        tmp_x = tmp_x.astype(np.float32) / 255.
                        tmp_y = np.zeros((len(self.classes))).astype(np.float32)
                        tmp_y[self.classes.index(i)] = 1.
                        x_batch.append(tmp_x)
                        y_pred.append(tmp_y)

                    except Exception as e:
                        print ("-----------------------------------------------------------------------------")
                        print ('>>> WARNING: fail handling ' +  file_list_by_class[i][idx[i]], e)
                        print ("-----------------------------------------------------------------------------")
                    
                    idx[i] += 1

            c = list(zip(x_batch, y_pred))
            random.shuffle(c)
            x_batch, y_pred = zip(*c)
            yield (np.array(x_batch), np.array(y_pred))
    

    def optimize(self, 
                 iteration, 
                 subdivition,
                 cost_tensor,
                 optimizer_tensor,
                 out_tensor, 
                 session,
                 saver, 
                 train_generator,
                 val_generator,
                 best_loss,
                 path_tosave_model='model/model1'):
        """[summary]
        
        Arguments:
            iteration {[type]} -- [description]
            subdivition {[type]} -- [description]
            cost_tensor {[type]} -- [description]
            optimizer_tensor {[type]} -- [description]
            out_tensor {[type]} -- [description]
        
        Keyword Arguments:
            train_batch_size {int} -- [description] (default: {32})
            val_batch_size {int} -- [description] (default: {50})
            path_tosave_model {str} -- [description] (default: {'model/model1'})
        """
        from sklearn.metrics import accuracy_score

        self.train_loss = []
        self.val_loss = []
        self.train_acc = []
        self.val_acc = []
        best_loss = best_loss
        
        for i in range(iteration):
            sign = "-"
            t_losses = []

            for j in range(subdivition):
                x_train, y_train = next(train_generator)
                feed_dict = {}
                feed_dict[self.input_placeholder] = x_train
                feed_dict[self.output_placeholder] = y_train
                session.run(optimizer_tensor, feed_dict)
                loss = session.run(cost_tensor, feed_dict)
                t_losses.append(loss)
                print ("> Train sub", j, 'loss : ', loss)
                
            x_val, y_val = next(val_generator)
            feed_dict = {}
            feed_dict[self.input_placeholder] = x_val
            feed_dict[self.output_placeholder] = y_val
            loss = session.run(cost_tensor, feed_dict)

            val_out = session.run(out_tensor, feed_dict)
            val_out = np.argmax(val_out, axis=1)
            y_val =  np.argmax(y_val, axis=1)
            val_acc = accuracy_score(val_out, y_val)
            
            t_loss = sum(t_losses) / (len(t_losses) + 0.0001)
            self.train_loss.append(t_loss)
            self.val_loss.append(loss)
                
            if best_loss > loss:
                best_loss = loss
                sign = "************* model saved"
                saver.save(session, path_tosave_model)
        
            print (">> epoch: ", i, "train loss: ", round(t_loss, 3), "val loss: ", round(loss, 3), "val acc: ", round(val_acc, 3), sign)


    



import cv2
import math
import random
import numpy as np 
import tensorflow as tf 
import qoalai.networks.densenet as densenet
from simple_tensor.tensor_operations import *
from simple_tensor.tensor_losses import mse_loss_mean, mse_loss_sum
from comdutils.file_utils import *
slim = tf.contrib.slim


class Similarity(): 
    def __init__(self, 
                 classes = ['normal', 'hard_damage', 'other'],
                 input_height = 300,
                 input_width = 300, 
                 input_channel = 3,
                 num_feature = 128): 

        self.input_height = input_height
        self.input_width = input_width
        self.input_channel = input_channel
        self.num_feature = num_feature
        self.classes = classes

        self.input_placeholder = tf.placeholder(tf.float32, shape=(None, self.input_height, self.input_width, self.input_channel))
        self.input_placeholder_pos = tf.placeholder(tf.float32, shape=(None, self.input_height, self.input_width, self.input_channel))
        self.input_placeholder_neg = tf.placeholder(tf.float32, shape=(None, self.input_height, self.input_width, self.input_channel))
        self.input_placeholder_other = tf.placeholder(tf.float32, shape=(None, self.input_height, self.input_width, self.input_channel))

    
    def build_net(self, input_tensor, is_training, reuse=None): 
        arg_scoope = densenet.densenet_arg_scope()
        out = None
        base_var_list = None
        
        with slim.arg_scope(arg_scoope):
            out = densenet.densenet121(inputs=input_tensor, 
                                       num_classes=1, 
                                       is_training=is_training,
                                       reuse = reuse)
            base_var_list = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)
        out = tf.keras.layers.Flatten()(out)
        
        out = tf.layers.Dense(units=128, use_bias=False, _reuse=reuse)(out)
        out = tf.nn.sigmoid(out)
    
        return out, base_var_list

    
    def calculate_loss(self, feature1, feature2, feature3, feature4): 
        self.pos_loss = mse_loss_mean(feature1, feature2)
        self.neg_loss = mse_loss_mean(feature1, feature3)
        self.other_loss = mse_loss_mean(feature1, feature4)
        self.loss = self.pos_loss - 1.3 * self.neg_loss - self.other_loss + 10.

    
    def batch_generator(self, batch_size, dataset_path, message):
        """Train Generator
        
        Arguments:
            batch_size {integer} -- the size of the batch
            image_name_list {list of string} -- the list of image name
        """
        file_list_by_class = {}
        idx = {}
        for i in self.classes:
            file_list_by_class[i] = get_filenames(dataset_path + i)
            random.shuffle(file_list_by_class[i])
            idx[i] = 0
        
        perclass_sample = int(batch_size/len(self.classes))

        print ("------------------------INFO IMAGES-------------------")
        print ("Image Folder: " + dataset_path)
        for i in self.classes:
            print ("Number of Image in " + str(i) + ": ", len(file_list_by_class[i]))
        print ("------------------------------------------------------")

        # Infinite loop.
        while True:
            x_batch = []
            x_pos = []
            x_neg = []
            x_other = []

            for i in self.classes:
                if i != 'other':
                    for j in range(perclass_sample):
                        if idx[i] >= len(file_list_by_class[i]):
                            random.shuffle(file_list_by_class[i])
                            print ("==>>> INFO: your " + message + " in class " + str(i) + " dataset is reshuffled again", idx[i])
                            idx[i] = 0
                        try:
                            tmp_x = cv2.imread(dataset_path + i + "/" + file_list_by_class[i][idx[i]])
                            tmp_x = cv2.cvtColor(tmp_x, cv2.COLOR_BGR2RGB)
                            tmp_x = cv2.resize(tmp_x, dsize=(self.input_width, self.input_height), interpolation=cv2.INTER_CUBIC)
                            tmp_x = tmp_x.astype(np.float32) / 255.
                            x_batch.append(tmp_x)
                            #print ("target: " + i + "/" + file_list_by_class[i][idx[i]])
                            
                            # ------------------------------------ #
                            # get the positive input image         #
                            # ------------------------------------ #
                            idx_pos= random.randint(0, len(file_list_by_class[i])-1)
                            tmp_x_pos = cv2.imread(dataset_path + i + "/" + file_list_by_class[i][idx_pos])
                            tmp_x_pos = cv2.cvtColor(tmp_x_pos, cv2.COLOR_BGR2RGB)
                            tmp_x_pos = cv2.resize(tmp_x_pos, dsize=(self.input_width, self.input_height), interpolation=cv2.INTER_CUBIC)
                            tmp_x_pos = tmp_x_pos.astype(np.float32) / 255.
                            x_pos.append(tmp_x_pos)
                            #print ("pos: " + i +  "/" + file_list_by_class[i][idx_pos])

                            # ------------------------------------ #
                            # get the negative input image         #
                            # ------------------------------------ #
                            the_list = list(file_list_by_class.keys())
                            the_list.remove(i)
                            if i != 'other':
                                the_list.remove('other')
                            tmp_class = random.choice(the_list)

                            idx_neg = random.randint(0, len(file_list_by_class[tmp_class])-1)
                            tmp_x_neg = cv2.imread(dataset_path + tmp_class + "/" + file_list_by_class[tmp_class][idx_neg])
                            tmp_x_neg = cv2.cvtColor(tmp_x_neg, cv2.COLOR_BGR2RGB)
                            tmp_x_neg = cv2.resize(tmp_x_neg, dsize=(self.input_width, self.input_height), interpolation=cv2.INTER_CUBIC)
                            tmp_x_neg = tmp_x_neg.astype(np.float32) / 255.
                            x_neg.append(tmp_x_neg)
                            #print ("neg: " + tmp_class +  "/" + file_list_by_class[tmp_class][idx_neg])

                            # ------------------------------------ #
                            # get the "other" input image          #
                            # ------------------------------------ #
                            idx_other = random.randint(0, len(file_list_by_class['other'])-1)
                            tmp_x_other = cv2.imread(dataset_path + 'other' + "/" + file_list_by_class['other'][idx_other])
                            tmp_x_other = cv2.cvtColor(tmp_x_other, cv2.COLOR_BGR2RGB)
                            tmp_x_other = cv2.resize(tmp_x_other, dsize=(self.input_width, self.input_height), interpolation=cv2.INTER_CUBIC)
                            tmp_x_other = tmp_x_other.astype(np.float32) / 255. 
                            x_other.append(tmp_x_other)
                            #print ("neg: " + 'other' + "/" + file_list_by_class['other'][idx_other])

                        except Exception as e:
                            print ("-----------------------------------------------------------------------------")
                            print ('>>> WARNING: fail handling ' +  file_list_by_class[i][idx[i]], e)
                            print ("-----------------------------------------------------------------------------")
                        
                        idx[i] += 1

            c = list(zip(x_batch, x_pos, x_neg, x_other))
            random.shuffle(c)
            x_batch, x_pos, x_neg, x_other = zip(*c)
            yield (np.array(x_batch), np.array(x_pos), np.array(x_neg), np.array(x_other))


    def optimize(self, 
                 iteration, 
                 subdivition,
                 cost_tensor,
                 optimizer_tensor,
                 session,
                 saver, 
                 train_generator,
                 val_generator,
                 best_loss,
                 path_tosave_model='model/model1'):
        """[summary]
        
        Arguments:
            iteration {[type]} -- [description]
            subdivition {[type]} -- [description]
            cost_tensor {[type]} -- [description]
            optimizer_tensor {[type]} -- [description]
            out_tensor {[type]} -- [description]
        
        Keyword Arguments:
            train_batch_size {int} -- [description] (default: {32})
            val_batch_size {int} -- [description] (default: {50})
            path_tosave_model {str} -- [description] (default: {'model/model1'})
        """

        self.train_loss = []
        self.val_loss = []
        best_loss = best_loss
        
        for i in range(iteration):
            sign = "-"
            t_losses = []

            for j in range(subdivition):
                x_train, x_train_pos, x_train_neg, x_train_other = next(train_generator)
                feed_dict = {}
                feed_dict[self.input_placeholder] = x_train
                feed_dict[self.input_placeholder_pos] = x_train_pos
                feed_dict[self.input_placeholder_neg] = x_train_neg
                feed_dict[self.input_placeholder_other] = x_train_other
                session.run(optimizer_tensor, feed_dict)
                loss = session.run(self.loss, feed_dict) 
                t_losses.append(loss)
                print ("> Train sub", j, 'loss : ', loss)
                
            x_val, y_val_pos, y_val_neg, y_val_other = next(val_generator)
            feed_dict = {}
            feed_dict[self.input_placeholder] = x_val
            feed_dict[self.input_placeholder_pos] = y_val_pos
            feed_dict[self.input_placeholder_neg] = y_val_neg
            feed_dict[self.input_placeholder_other] = y_val_other
            loss, p_loss, n_loss, o_loss = session.run([self.loss, self.pos_loss, self.neg_loss, self.other_loss], feed_dict)
            
            t_loss = sum(t_losses) / (len(t_losses) + 0.0001)
            self.train_loss.append(t_loss)
            self.val_loss.append(loss)
                
            if best_loss > loss:
                best_loss = loss
                sign = "************* model saved"
                saver.save(session, path_tosave_model)
        
            print (">> epoch: ", i, "train loss: ", round(t_loss, 3), "val loss: ", round(loss, 3), \
                "p loss: ", round(p_loss, 3), "n loss: ", round(n_loss, 3), "o loss: ", round(o_loss,3), sign)  



'''
    File name: network.py
    Author: [DS Team]
    Date created: / /2020
    Date last modified: 
    Python Version: >= 3.5
    qoalai version: v0.4
    License: MIT License
    Maintainer: [DS Team]
'''

import os
import cv2
import math
import numpy as np
from os import walk
import tensorflow as tf 
from qoalai.tensor_operations import *



class YoloNetwork(object): 
    def __init__(self,
                    num_class,
                    is_multilabel,
                    multilabel_dict,
                    anchor, 
                    dropout_val = 0.85,
                    threshold = 0.5,
                    leaky_relu_alpha = 0.1,
                    add_modsig_toshape = True,
                    convert_to_tflite=False): 

        self.num_class = num_class
        self.is_multilabel = is_multilabel
        self.multilabel_dict = multilabel_dict 
        self.anchor = anchor
        self.dropout_val = 0.8
        self.threshold = threshold
        self.leaky_relu_alpha = leaky_relu_alpha
        self.add_modsig_toshape = add_modsig_toshape
        self.lite = convert_to_tflite

        self.FILTER_128 = 128
        self.FILTER_256 = 256
        self.FILTER_512 = 512
        self.FILTER_1024 = 1024 

    
    def fixed_padding(self, inputs, 
                        kernel_size, 
                        data_format):
        """ResNet implementation of fixed padding.
        Pads the input along the spatial dimensions independently of input size.

        Args:
            inputs: Tensor input to be padded.
            kernel_size: The kernel to be used in the conv2d or max_pool2d.
            data_format: The input format.
        Returns:
            A tensor with the same format as the input.
        """
        pad_total = kernel_size - 1
        pad_beg = pad_total // 2
        pad_end = pad_total - pad_beg

        if data_format == 'channels_first':
            padded_inputs = tf.pad(inputs, [[0, 0], [0, 0],
                                            [pad_beg, pad_end],
                                            [pad_beg, pad_end]])
        else:
            padded_inputs = tf.pad(inputs, [[0, 0], [pad_beg, pad_end],
                                            [pad_beg, pad_end], [0, 0]])
        return padded_inputs 

        
    def darknet53_residual_block(self, inputs, 
                                    filters, 
                                    training, 
                                    data_format, 
                                    stride=1, 
                                    name='res'):
        """[summary]
        Arguments:
            inputs {[tensor]} -- input tensor
            filters {[tensor]} -- the filter tensor
            training {[bool]} -- the phase, training or not
            data_format {[string]} -- channel first or not
        
        Keyword Arguments:
            stride {int} -- [the strides] (default: {1})
            name {str} -- [the additional name for all tensors in this block] (default: {'res'})
        
        Returns:
            [type] -- [description]
        """
        shortcut = inputs
        
        inputs = new_conv2d_layer(input=inputs, 
                        filter_shape=[1, 1, inputs.get_shape().as_list()[-1], filters], 
                        name = name + '_input_conv1', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',  
                        padding=('SAME' if stride == 1 else 'VALID'), 
                        strides=[1, stride, stride, 1],  
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)
        
        inputs = new_conv2d_layer(input=(inputs if stride == 1 else self.fixed_padding(inputs, 3, 'channels_last')), 
                        filter_shape=[3, 3, inputs.get_shape().as_list()[-1], shortcut.get_shape().as_list()[-1]], 
                        name = name + '_input_conv2', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding=('SAME' if stride == 1 else 'VALID'), 
                        strides=[1, stride, stride, 1], 
                        data_type=tf.float32, 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)

        inputs += shortcut
        return inputs


    def darknet53(self, inputs, training, data_format, network_type):
        """[summary]
        
        Arguments:
            inputs {tensor} -- the input tensor
            training {bool} -- the phase, training or not
            data_format {string} -- channel_first or not
        
        Returns:
            [type] -- [description]
        """
        inputs = new_conv2d_layer(input=inputs, 
                        filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 32], 
                        name = 'main_input_conv1', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding='SAME', 
                        strides=[1, 1, 1, 1], 
                        data_type=tf.float32, 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)

        inputs = new_conv2d_layer(input=self.fixed_padding(inputs, 3, 'channels_last'), 
                        filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 64], 
                        name = 'main_input_conv2', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding='VALID', 
                        strides=[1, 2, 2, 1], 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)

        inputs = self.darknet53_residual_block(inputs=inputs, 
                                            filters=32, 
                                            training=training,
                                            data_format=data_format, 
                                            name='res1')

        inputs = new_conv2d_layer(input=self.fixed_padding(inputs, 3, 'channels_last'), 
                        filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 128], 
                        name = 'main_input_conv3', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding='VALID', 
                        strides=[1, 2, 2, 1], 
                        data_type=tf.float32, 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)

        for i in range(2):
            inputs = self.darknet53_residual_block(inputs, 
                                                filters=64,
                                                training=training,
                                                data_format=data_format, 
                                                name='res' + str(i+1))
            
        inputs = new_conv2d_layer(input=self.fixed_padding(inputs, 3, 'channels_last'), 
                        filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 256], 
                        name = 'main_input_conv4', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding='VALID', 
                        strides=[1, 2, 2, 1],  
                        data_type=tf.float32, 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)

        return_vars = None
        if network_type == 'special':
            return_vars = tf.global_variables(scope='yolo_v3_model')
        elif network_type == 'very_small':
            return_vars = tf.global_variables(scope='yolo_v3_model')

        for i in range(8):
            inputs = self.darknet53_residual_block(inputs, 
                                                filters=self.FILTER_128,
                                                training=training,
                                                data_format=data_format, 
                                                name='res' + str(i+3))

        route1 = inputs
        inputs = new_conv2d_layer(input=self.fixed_padding(inputs, 3, 'channels_last'), 
                        filter_shape=[3, 3, inputs.get_shape().as_list()[-1], self.FILTER_512], 
                        name = 'main_input_conv5', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding='VALID', 
                        strides=[1, 2, 2, 1], 
                        data_type=tf.float32, 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)
        
        for i in range(8):
            inputs = self.darknet53_residual_block(inputs, 
                                                filters=self.FILTER_256,
                                                training=training,
                                                data_format=data_format, 
                                                name='res' + str(i+11))

        route2 = inputs
        inputs = new_conv2d_layer(input=self.fixed_padding(inputs, 3, 'channels_last'), 
                        filter_shape=[3, 3, inputs.get_shape().as_list()[-1], self.FILTER_1024], 
                        name = 'main_input_conv6', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU', 
                        padding='VALID', 
                        strides=[1, 2, 2, 1],  
                        data_type=tf.float32, 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)

        for i in range(4):
            inputs = self.darknet53_residual_block(inputs, filters=self.FILTER_512,
                                            training=training,
                                            data_format=data_format, name='res' + str(i+19))
        return route1, route2, inputs, return_vars


    def yolo_convolution_block(self, inputs, filters, training, data_format):
        """[summary]
        
        Arguments:
            inputs {[type]} -- [description]
            filters {[type]} -- [description]
            training {[type]} -- [description]
            data_format {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        inputs = new_conv2d_layer(input=inputs, 
                        filter_shape=[1, 1, inputs.get_shape().as_list()[-1], filters], 
                        name = 'main_input_conv7', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding='SAME', 
                        strides=[1, 1, 1, 1],
                        data_type=tf.float32, 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)
    
        inputs = new_conv2d_layer(input=inputs, 
                        filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 2*filters], 
                        name = 'main_input_conv8', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding='SAME', 
                        strides=[1, 1, 1, 1],
                        data_type=tf.float32, 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)
        
        inputs = new_conv2d_layer(input=inputs, 
                        filter_shape=[1, 1, inputs.get_shape().as_list()[-1], filters], 
                        name = 'main_input_conv9', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding='SAME', 
                        strides=[1, 1, 1, 1],
                        data_type=tf.float32,   
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)
        
        inputs = new_conv2d_layer(input=inputs, 
                        filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 2*filters], 
                        name = 'main_input_conv10', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding='SAME', 
                        strides=[1, 1, 1, 1],
                        data_type=tf.float32, 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)
        
        inputs = new_conv2d_layer(input=inputs, 
                        filter_shape=[1, 1, inputs.get_shape().as_list()[-1], filters], 
                        name = 'main_input_conv11', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding='SAME', 
                        strides=[1, 1, 1, 1], 
                        data_type=tf.float32, 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)

        route = inputs
        inputs = new_conv2d_layer(input=inputs, 
                        filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 2*filters], 
                        name = 'main_input_conv12', 
                        dropout_val= self.dropout_val, 
                        activation = 'LRELU',
                        padding='SAME', 
                        strides=[1, 1, 1, 1],
                        data_type=tf.float32, 
                        is_training=training,
                        use_bias=False,
                        use_batchnorm=True)
                    
        return route, inputs

    
    def yolo_layer(self, inputs, n_classes, 
                    anchors, img_size, data_format):
        """Creates Yolo final detection layer.

        Detects boxes with respect to anchors.

        Args:
            inputs: Tensor input.
            n_classes: Number of labels.
            anchors: A list of anchor sizes.
            img_size: The input size of the model.
            data_format: The input format.

        Returns:
            Tensor output.
        """
        # ----------------------------- #
        # get the num of anchor         #
        # get the grid shape (hxw)      #
        # ----------------------------- #
        n_anchors = len(anchors)
        shape = inputs.get_shape().as_list()
        grid_shape = shape[2:4] if data_format == 'channels_first' else shape[1:3]

        # ----------------------------- #
        # reshape the tensor to:        #
        # batch x (num of anchor . h . w) x (5 + num of classes)
        # strides -> int(the image input / grid)
        # strides -> horizontal and vertical stride 
        # ----------------------------- #
        if data_format == 'channels_first':
            inputs = tf.transpose(inputs, [0, 2, 3, 1])
        inputs = tf.reshape(inputs, [-1, n_anchors * grid_shape[0] * grid_shape[1],
                                    5 + n_classes])

        strides = (img_size[0] // grid_shape[0], img_size[1] // grid_shape[1])

        # ----------------------------- #
        # split the tensor to           #
        # center, shape, confidence, class
        # center: SILL RELATIVE to the GRID
        # create an x ofset tensor based on grid position
        # create an y ofset tensor based on grid position
        # center = (ofset tensor + center) * stride
        # ----------------------------- #
        box_centers, box_shapes, confidence, classes = \
                    tf.split(inputs, [2, 2, 1, n_classes], axis=-1)

        x = tf.range(grid_shape[0], dtype=tf.float32)
        y = tf.range(grid_shape[1], dtype=tf.float32)
        x_offset, y_offset = tf.meshgrid(x, y)
        x_offset = tf.reshape(x_offset, (-1, 1))
        y_offset = tf.reshape(y_offset, (-1, 1))
        x_y_offset = tf.concat([x_offset, y_offset], axis=-1)
        x_y_offset = tf.tile(x_y_offset, [1, n_anchors])
        x_y_offset = tf.reshape(x_y_offset, [1, -1, 2])
        box_centers = tf.nn.sigmoid(box_centers)
        box_centers = (box_centers + x_y_offset) * strides

        # ----------------------------- #
        # shape = exp(shape) * anchor
        # box confidence = sigmoid(conf)
        # classes = sigmoid(classes)
        # final -> concat all
        # ----------------------------- #
        anchors = tf.tile(anchors, [grid_shape[0] * grid_shape[1], 1])
        if self.add_modsig_toshape:
            box_shapes = 6 /(1 + tf.exp(-0.2 * box_shapes)) - 3
        box_shapes = tf.exp(box_shapes) * tf.to_float(anchors)
        confidence = tf.nn.sigmoid(confidence)
        classes = tf.nn.sigmoid(classes)
        inputs = tf.concat([box_centers, box_shapes,
                            confidence, classes], axis=-1)
        return inputs


    def upsample(self, inputs, out_shape, data_format):
        """Upsamples to `out_shape` using nearest neighbor interpolation.
        
        Arguments:
            inputs {[type]} -- [description]
            out_shape {[type]} -- [description]
            data_format {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        if data_format == 'channels_first':
            inputs = tf.transpose(inputs, [0, 2, 3, 1])
            new_height = out_shape[3]
            new_width = out_shape[2]
        else:
            new_height = out_shape[2]
            new_width = out_shape[1]

        inputs = tf.image.resize_nearest_neighbor(inputs, (new_height, new_width)) #tf.compat.v1.image.resize_nearest_neighbor(inputs, (new_height, new_width))
        if data_format == 'channels_first':
            inputs = tf.transpose(inputs, [0, 3, 1, 2])
        return inputs


    def build_boxes(self, inputs):
        """Computes top left and bottom right points of the boxes.
        
        Arguments:
            inputs {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        center_x, center_y, width, height, confidence, classes = \
            tf.split(inputs, [1, 1, 1, 1, 1, -1], axis=-1)

        if self.num_class == 1:
            classes = tf.nn.sigmoid(classes)
        else:
            classes = tf.nn.softmax(classes)

        top_left_x = center_x - width / 2
        top_left_y = center_y - height / 2
        bottom_right_x = center_x + width / 2
        bottom_right_y = center_y + height / 2 
        
        if self.lite:
            '''
            boxes = tf.concat([top_left_x, top_left_y,
                        bottom_right_x, bottom_right_y,
                        confidence, classes], axis=-1)
            tmp_boxes = boxes[:, :, :-(self.num_class + 1)]
            self.selected_indices = tf.image.non_max_suppression(tmp_boxes[0, :],
                                                                tf.squeeze(confidence[0, :]),
                                                                max_output_size = 15,
                                                                iou_threshold=0.5,
                                                                score_threshold=float('-inf'),
                                                                name=None)
            self.selected_boxes = tf.gather(boxes, self.selected_indices, axis=1)
            
            results = {} 
            results['detection_boxes'] = tf.concat([top_left_x, top_left_y, bottom_right_x, bottom_right_y], axis = -1)
            results['detection_classes'] = classes
            results['detection_scores'] = confidence
            return results
            '''
            boxes = tf.concat([top_left_x, top_left_y,
                        bottom_right_x, bottom_right_y,
                        confidence, classes], axis=-1)
            return boxes

        else: 
            selected_mask = tf.math.greater(confidence, tf.convert_to_tensor(np.array(self.threshold), tf.float32))
            total_num = tf.reduce_sum(tf.cast(selected_mask, tf.float32))
            
            #----------------------------------#
            # Remove some bbox with confidence lower than thd       
            #----------------------------------#
            confidence = tf.boolean_mask(confidence, selected_mask, axis=None)
            classes = tf.boolean_mask(classes, tf.squeeze(selected_mask, 2), axis=0)
            top_left_x = tf.boolean_mask(top_left_x, selected_mask, axis=None)
            top_left_y = tf.boolean_mask(top_left_y, selected_mask, axis=None)
            bottom_right_y = tf.boolean_mask(bottom_right_y, selected_mask, axis=None)
            bottom_right_x = tf.boolean_mask(bottom_right_x, selected_mask, axis=None)
            
            confidence = tf.reshape(tensor=confidence, shape=(-1, total_num, 1))
            classes = tf.reshape(tensor=classes, shape=(-1, total_num, self.num_class))
            top_left_x = tf.reshape(tensor=top_left_x, shape=(-1, total_num, 1))
            top_left_y = tf.reshape(tensor=top_left_y, shape=(-1, total_num, 1))
            bottom_right_x = tf.reshape(tensor=bottom_right_x, shape=(-1, total_num, 1))
            bottom_right_y = tf.reshape(tensor=bottom_right_y, shape=(-1, total_num, 1))

            boxes = tf.concat([top_left_x, top_left_y,
                        bottom_right_x, bottom_right_y,
                        confidence, classes], axis=-1)
            return boxes


    def build_yolov3_net(self, inputs, network_type, is_training):
        """method for building yolo-v3 network
        
        Returns:
            [type] -- [description]
        """
        model_size = (416, 416)
        data_format = 'channels_last'

        self.network_type = network_type
        if self.network_type == 'special': 
            self.FILTER_128 = 32
            self.FILTER_256 = 32
            self.FILTER_512 = 32
            self.FILTER_1024 = 32
        elif self.network_type == 'very_small': 
            self.FILTER_128 = 64
            self.FILTER_256 = 64
            self.FILTER_512 = 64
            self.FILTER_1024 = 64

        if self.network_type == 'special' or \
            self.network_type == 'very_small':
            route1, route2, inputs, self.yolo_special_vars = darknet53(inputs, 
                                                                        training=is_training,
                                                                        data_format=data_format, 
                                                                        network_type=network_type)
            self.yolo_very_small_vars = self.yolo_special_vars
        else:
            route1, route2, inputs, _ = self.darknet53(inputs, 
                                                    training=is_training,
                                                    data_format=data_format, 
                                                    network_type=network_type)

        self.YOLO_CONV_FILTERS = {}
        if self.network_type == 'big':
            self.YOLO_CONV_FILTERS['a'] = 512
            self.YOLO_CONV_FILTERS['b'] = 256
            self.YOLO_CONV_FILTERS['c'] = 128
        elif network_type == 'medium':
            self.YOLO_CONV_FILTERS['a'] = 512
            self.YOLO_CONV_FILTERS['b'] = 128
            self.YOLO_CONV_FILTERS['c'] = 128
        elif network_type == 'small':
            self.YOLO_CONV_FILTERS['a'] = 256
            self.YOLO_CONV_FILTERS['b'] = 128
            self.YOLO_CONV_FILTERS['c'] = 128
        elif network_type == 'very_small':
            self.YOLO_CONV_FILTERS['a'] = 128
            self.YOLO_CONV_FILTERS['b'] = 64
            self.YOLO_CONV_FILTERS['c'] = 64
        elif network_type == 'special':
            self.YOLO_CONV_FILTERS['a'] = 64
            self.YOLO_CONV_FILTERS['b'] = 32
            self.YOLO_CONV_FILTERS['c'] = 32

        self.yolo_small_vars = tf.global_variables(scope='yolo_v3_model')
        route, inputs = self.yolo_convolution_block(inputs, 
                                                filters=self.YOLO_CONV_FILTERS['a'], 
                                                training=is_training,
                                                data_format=data_format)
        inputs_detect1 = inputs
        inputs = new_conv2d_layer(input=route, 
                    filter_shape=[1, 1, route.get_shape().as_list()[-1], 256], 
                    name = 'main_input_conv13', 
                    dropout_val= self.dropout_val, 
                    activation = 'LRELU',
                    padding='SAME', 
                    strides=[1, 1, 1, 1], 
                    data_type=tf.float32, 
                    is_training=is_training,
                    use_bias=False,
                    use_batchnorm=True)

        upsample_size = route2.get_shape().as_list()
        inputs = self.upsample(inputs, 
                            out_shape=upsample_size,
                            data_format=data_format)
        axis = 3
        inputs = tf.concat([inputs, route2], axis=axis)

        self.yolo_medium_vars = tf.global_variables(scope='yolo_v3_model')
        route, inputs = self.yolo_convolution_block(inputs, 
                                                filters=self.YOLO_CONV_FILTERS['b'],  
                                                training=is_training,
                                                data_format=data_format)

        inputs_detect2 = inputs
        inputs = new_conv2d_layer(input=route, 
                                    filter_shape=[1, 1, route.get_shape().as_list()[-1], 128], 
                                    name = 'main_input_conv14', 
                                    dropout_val= self.dropout_val, 
                                    activation = 'LRELU',
                                    padding='SAME', 
                                    strides=[1, 1, 1, 1],
                                    data_type=tf.float32,
                                    is_training=is_training,
                                    use_bias=False,
                                    use_batchnorm=True)
        upsample_size = route1.get_shape().as_list()

        inputs = self.upsample(inputs, 
                            out_shape=upsample_size,
                            data_format=data_format)
        inputs = tf.concat([inputs, route1], axis=axis)
        route, inputs = self.yolo_convolution_block(inputs, 
                                                filters=self.YOLO_CONV_FILTERS['c'], 
                                                training=is_training,
                                                data_format=data_format)

        self.yolo_big_vars = tf.global_variables(scope='yolo_v3_model')
        self.yolo_vars = tf.global_variables(scope='yolo_v3_model')

        inputs_detect3 = inputs

        if self.is_multilabel: 
            total_properties = 6 #sum(self.multilabel_dict.values)
            OUTPUT_DEPTH = int(len(self.anchor)/3 * (5 + self.num_class + total_properties))
        else: 
            OUTPUT_DEPTH = int(len(self.anchor)/3 * (5 + self.num_class))
        
        self.detect1 = tf.layers.conv2d(inputs_detect1, 
                                    filters=OUTPUT_DEPTH,
                                    kernel_size=1, 
                                    strides=1, 
                                    use_bias=True,
                                    data_format=data_format)
     
        self.detect2 = tf.layers.conv2d(inputs_detect2, 
                                    filters=OUTPUT_DEPTH,
                                    kernel_size=1, 
                                    strides=1, 
                                    use_bias=True,
                                    data_format=data_format)

        self.detect3 = tf.layers.conv2d(inputs_detect3, 
                                    filters=OUTPUT_DEPTH,
                                    kernel_size=1, 
                                    strides=1, 
                                    use_bias=True,
                                    data_format=data_format)

        """  
        self.detect1 = new_conv2d_layer(input=inputs_detect1, 
                                    filter_shape=[1, 1, inputs_detect1.get_shape().as_list()[-1], OUTPUT_DEPTH], 
                                    name = 'detect1_conv', 
                                    dropout_val= self.dropout_val, 
                                    activation = 'None',
                                    padding='VALID', 
                                    strides=[1, 1, 1, 1],
                                    is_training=is_training,
                                    use_bias=True,
                                    use_batchnorm=False)
        self.detect2 = new_conv2d_layer(input=inputs_detect2, 
                                    filter_shape=[1, 1, inputs_detect2.get_shape().as_list()[-1], OUTPUT_DEPTH], 
                                    name = 'detect2_conv', 
                                    dropout_val= self.dropout_val, 
                                    activation = 'None',
                                    padding='VALID', 
                                    strides=[1, 1, 1, 1],
                                    is_training=is_training,
                                    use_bias=True,
                                    use_batchnorm=False)
        self.detect3 = new_conv2d_layer(input=inputs_detect3, 
                                    filter_shape=[1, 1, inputs_detect3.get_shape().as_list()[-1], OUTPUT_DEPTH], 
                                    name = 'detect2_conv', 
                                    dropout_val= self.dropout_val, 
                                    activation = 'None',
                                    padding='VALID', 
                                    strides=[1, 1, 1, 1],
                                    is_training=is_training,
                                    use_bias=True,
                                    use_batchnorm=False)
        """
        self.output_list = [self.detect1, self.detect2, self.detect3]
        
        combine_box1 = self.yolo_layer(self.detect1, 
                                n_classes=self.num_class,
                                anchors=self.anchor[6:9],
                                img_size=model_size,
                                data_format=data_format)
        combine_box2 = self.yolo_layer(self.detect2, 
                                n_classes=self.num_class,
                                anchors=self.anchor[3:6],
                                img_size=model_size,
                                data_format=data_format)
        combine_box3 = self.yolo_layer(self.detect3, 
                                n_classes=self.num_class,
                                anchors=self.anchor[0:3],
                                img_size=model_size,
                                data_format=data_format)
        
        if self.lite: 
            inputs = combine_box1
        else: 
            inputs = tf.concat([combine_box1, combine_box2, combine_box3], axis=1)
        self.boxes_dicts = self.build_boxes(inputs)


    def build_yolov3_net2(self, inputs, network_type, is_training):
        """method for building yolo-v3 network
        
        Returns:
            [type] -- [description]
        """
        print ("=====================================")
        print ("Using Old Version")
        print ("=====================================")
        model_size = (416, 416)
        max_output_size = 10
        data_format = 'channels_last'


        # -------------------------------------------- #
        # Function for giving hard padding             #
        # It is applied on original yolo network       #
        # -------------------------------------------- #
        def fixed_padding(inputs, 
                          kernel_size, 
                          data_format):
            """ResNet implementation of fixed padding.
            Pads the input along the spatial dimensions independently of input size.

            Args:
                inputs: Tensor input to be padded.
                kernel_size: The kernel to be used in the conv2d or max_pool2d.
                data_format: The input format.
            Returns:
                A tensor with the same format as the input.
            """
            pad_total = kernel_size - 1
            pad_beg = pad_total // 2
            pad_end = pad_total - pad_beg

            if data_format == 'channels_first':
                padded_inputs = tf.pad(inputs, [[0, 0], [0, 0],
                                                [pad_beg, pad_end],
                                                [pad_beg, pad_end]])
            else:
                padded_inputs = tf.pad(inputs, [[0, 0], [pad_beg, pad_end],
                                                [pad_beg, pad_end], [0, 0]])
            return padded_inputs
        
        # --------------------------------------- #
        # For building darknet 53 residual block  #
        # --------------------------------------- #           
        def darknet53_residual_block(inputs, 
                                     filters, 
                                     training, 
                                     data_format, 
                                     stride=1, 
                                     name='res'):
            """[summary]
            Arguments:
                inputs {[tensor]} -- input tensor
                filters {[tensor]} -- the filter tensor
                training {[bool]} -- the phase, training or not
                data_format {[string]} -- channel first or not
            
            Keyword Arguments:
                stride {int} -- [the strides] (default: {1})
                name {str} -- [the additional name for all tensors in this block] (default: {'res'})
            
            Returns:
                [type] -- [description]
            """
            shortcut = inputs
            
            inputs = new_conv2d_layer(input=inputs, 
                            filter_shape=[1, 1, inputs.get_shape().as_list()[-1], filters], 
                            name = name + '_input_conv1', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU', 
                            lrelu_alpha= self.leaky_relu_alpha, 
                            padding=('SAME' if stride == 1 else 'VALID'), 
                            strides=[1, stride, stride, 1],  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)
            
            inputs = new_conv2d_layer(input=(inputs if stride == 1 else fixed_padding(inputs, 3, 'channels_last')), 
                            filter_shape=[3, 3, inputs.get_shape().as_list()[-1], shortcut.get_shape().as_list()[-1]], 
                            name = name + '_input_conv2', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha, 
                            padding=('SAME' if stride == 1 else 'VALID'), 
                            strides=[1, stride, stride, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)

            inputs += shortcut
            return inputs
        
        # ------------------------------------- #
        # function for building main darknet 53 #
        # ------------------------------------- #
        def darknet53(inputs, training, data_format, network_type):
            """[summary]
            
            Arguments:
                inputs {tensor} -- the input tensor
                training {bool} -- the phase, training or not
                data_format {string} -- channel_first or not
            
            Returns:
                [type] -- [description]
            """
            inputs = new_conv2d_layer(input=inputs, 
                            filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 32], 
                            name = 'main_input_conv1', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha,  
                            padding='SAME', 
                            strides=[1, 1, 1, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)

            inputs = new_conv2d_layer(input=fixed_padding(inputs, 3, 'channels_last'), 
                            filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 64], 
                            name = 'main_input_conv2', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha,  
                            padding='VALID', 
                            strides=[1, 2, 2, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)

            inputs = darknet53_residual_block(inputs, 
                                              filters=32, 
                                              training=training,
                                              data_format=data_format, 
                                              name='res1')

            inputs = new_conv2d_layer(input=fixed_padding(inputs, 3, 'channels_last'), 
                            filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 128], 
                            name = 'main_input_conv3', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha, 
                            padding='VALID', 
                            strides=[1, 2, 2, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)

            for i in range(2):
                inputs = darknet53_residual_block(inputs, 
                                                  filters=64,
                                                  training=training,
                                                  data_format=data_format, 
                                                  name='res' + str(i+1))
                
            inputs = new_conv2d_layer(input=fixed_padding(inputs, 3, 'channels_last'), 
                            filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 256], 
                            name = 'main_input_conv4', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha, 
                            padding='VALID', 
                            strides=[1, 2, 2, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)

            filter_128 = 128
            filter_256 = 256
            filter_512 = 512
            filter_1024 = 1024
            return_vars = None
            if network_type == 'special':
                filter_128 = 32
                filter_256 = 32
                filter_512 = 32
                filter_1024 = 32
                return_vars = tf.global_variables(scope='yolo_v3_model')
            elif network_type == 'very_small':
                filter_128 = 64
                filter_256 = 64
                filter_512 = 64
                filter_1024 = 64
                return_vars = tf.global_variables(scope='yolo_v3_model')

            for i in range(8):
                inputs = darknet53_residual_block(inputs, 
                                                  filters=filter_128,
                                                  training=training,
                                                  data_format=data_format, 
                                                  name='res' + str(i+3))

            route1 = inputs
            inputs = new_conv2d_layer(input=fixed_padding(inputs, 3, 'channels_last'), 
                            filter_shape=[3, 3, inputs.get_shape().as_list()[-1], filter_512], 
                            name = 'main_input_conv5', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha,
                            padding='VALID', 
                            strides=[1, 2, 2, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)
            
            for i in range(8):
                inputs = darknet53_residual_block(inputs, 
                                                  filters=filter_256,
                                                  training=training,
                                                  data_format=data_format, 
                                                  name='res' + str(i+11))

            route2 = inputs
            inputs = new_conv2d_layer(input=fixed_padding(inputs, 3, 'channels_last'), 
                            filter_shape=[3, 3, inputs.get_shape().as_list()[-1], filter_1024], 
                            name = 'main_input_conv6', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha, 
                            padding='VALID', 
                            strides=[1, 2, 2, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)

            for i in range(4):
                inputs = darknet53_residual_block(inputs, filters=filter_512,
                                                training=training,
                                                data_format=data_format, name='res' + str(i+19))
            return route1, route2, inputs, return_vars
        
        #---------------------------------------- 
        def yolo_convolution_block(inputs, filters, training, data_format):
            """[summary]
            
            Arguments:
                inputs {[type]} -- [description]
                filters {[type]} -- [description]
                training {[type]} -- [description]
                data_format {[type]} -- [description]
            
            Returns:
                [type] -- [description]
            """
            inputs = new_conv2d_layer(input=inputs, 
                            filter_shape=[1, 1, inputs.get_shape().as_list()[-1], filters], 
                            name = 'main_input_conv7', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha,
                            padding='SAME', 
                            strides=[1, 1, 1, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)
        
            inputs = new_conv2d_layer(input=inputs, 
                            filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 2*filters], 
                            name = 'main_input_conv8', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha,
                            padding='SAME', 
                            strides=[1, 1, 1, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)
            
            inputs = new_conv2d_layer(input=inputs, 
                            filter_shape=[1, 1, inputs.get_shape().as_list()[-1], filters], 
                            name = 'main_input_conv9', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha,
                            padding='SAME', 
                            strides=[1, 1, 1, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)
            
            inputs = new_conv2d_layer(input=inputs, 
                            filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 2*filters], 
                            name = 'main_input_conv10', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha, 
                            padding='SAME', 
                            strides=[1, 1, 1, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)
            
            inputs = new_conv2d_layer(input=inputs, 
                            filter_shape=[1, 1, inputs.get_shape().as_list()[-1], filters], 
                            name = 'main_input_conv11', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha, 
                            padding='SAME', 
                            strides=[1, 1, 1, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)

            route = inputs
            inputs = new_conv2d_layer(input=inputs, 
                            filter_shape=[3, 3, inputs.get_shape().as_list()[-1], 2*filters], 
                            name = 'main_input_conv12', 
                            dropout_val= self.dropout_val, 
                            activation = 'LRELU',
                            lrelu_alpha=self.leaky_relu_alpha, 
                            padding='SAME', 
                            strides=[1, 1, 1, 1],
                            data_type=tf.float32,  
                            is_training=training,
                            use_bias=False,
                            use_batchnorm=True)
                        
            return route, inputs
        
        #-----------------------------------#
        def yolo_layer(inputs, n_classes, anchors, img_size, data_format):
            """Creates Yolo final detection layer.

            Detects boxes with respect to anchors.

            Args:
                inputs: Tensor input.
                n_classes: Number of labels.
                anchors: A list of anchor sizes.
                img_size: The input size of the model.
                data_format: The input format.

            Returns:
                Tensor output.
            """
            
            n_anchors = len(anchors)
            shape = inputs.get_shape().as_list()
            grid_shape = shape[2:4] if data_format == 'channels_first' else shape[1:3]
            if data_format == 'channels_first':
                inputs = tf.transpose(inputs, [0, 2, 3, 1])

            strides = (img_size[0] // grid_shape[0], img_size[1] // grid_shape[1])

            if self.is_multilabel: 
                inputs = tf.reshape(inputs, [-1, n_anchors * grid_shape[0] * grid_shape[1], 5 + n_classes + sum(self.multilabel_dict.values())])
                box_centers, box_shapes, confidence, classes = tf.split(inputs, [2, 2, 1, n_classes + sum(self.multilabel_dict.values())], axis=-1)
            else: 
                inputs = tf.reshape(inputs, [-1, n_anchors * grid_shape[0] * grid_shape[1], 5 + n_classes])
                box_centers, box_shapes, confidence, classes = tf.split(inputs, [2, 2, 1, n_classes], axis=-1)

            x = tf.range(grid_shape[0], dtype=tf.float32)
            y = tf.range(grid_shape[1], dtype=tf.float32)
            x_offset, y_offset = tf.meshgrid(x, y)
            x_offset = tf.reshape(x_offset, (-1, 1))
            y_offset = tf.reshape(y_offset, (-1, 1))
            x_y_offset = tf.concat([x_offset, y_offset], axis=-1)
            x_y_offset = tf.tile(x_y_offset, [1, n_anchors])
            x_y_offset = tf.reshape(x_y_offset, [1, -1, 2])
            box_centers = tf.nn.sigmoid(box_centers)
            box_centers = (box_centers + x_y_offset) * strides

            anchors = tf.tile(anchors, [grid_shape[0] * grid_shape[1], 1])
            if self.add_modsig_toshape:
                box_shapes = 6 /(1 + tf.exp(-0.2 * box_shapes)) - 3
            box_shapes = tf.exp(box_shapes) * tf.to_float(anchors)
            confidence = tf.nn.sigmoid(confidence)
            inputs = tf.concat([box_centers, box_shapes,
                                confidence, classes], axis=-1)

            return inputs

        #---------------------------------------------#
        def upsample(inputs, out_shape, data_format):
            """Upsamples to `out_shape` using nearest neighbor interpolation.
            
            Arguments:
                inputs {[type]} -- [description]
                out_shape {[type]} -- [description]
                data_format {[type]} -- [description]
            
            Returns:
                [type] -- [description]
            """
            if data_format == 'channels_first':
                inputs = tf.transpose(inputs, [0, 2, 3, 1])
                new_height = out_shape[3]
                new_width = out_shape[2]
            else:
                new_height = out_shape[2]
                new_width = out_shape[1]

            inputs = tf.image.resize_nearest_neighbor(inputs, (new_height, new_width))
            if data_format == 'channels_first':
                inputs = tf.transpose(inputs, [0, 3, 1, 2])
            return inputs
        
        #------------------------------------------------#
        def build_boxes(inputs):
            """Computes top left and bottom right points of the boxes.
            
            Arguments:
                inputs {[type]} -- [description]
            
            Returns:
                [type] -- [description]
            """
            center_x, center_y, width, height, confidence, classes_all = \
                tf.split(inputs, [1, 1, 1, 1, 1, -1], axis=-1)

            additional_class = []
            if self.is_multilabel: 
                classes = classes_all[:, :, :self.num_class]
                base = self.num_class

                for i in self.multilabel_dict.values(): 
                    additional_class.append(tf.nn.softmax(classes_all[:, :, base:base + i]))
                    base = base + i
            else: 
                classes = classes_all

            if self.num_class == 1:
                classes = tf.nn.sigmoid(classes)
            else:
                classes = tf.nn.softmax(classes)

            top_left_x = center_x - width / 2
            top_left_y = center_y - height / 2
            bottom_right_x = center_x + width / 2
            bottom_right_y = center_y + height / 2 
            
            if self.lite:
                '''
                boxes = tf.concat([top_left_x, top_left_y,
                            bottom_right_x, bottom_right_y,
                            confidence, classes], axis=-1)
                tmp_boxes = boxes[:, :, :-(self.num_class + 1)]
                self.selected_indices = tf.image.non_max_suppression(tmp_boxes[0, :],
                                                                    tf.squeeze(confidence[0, :]),
                                                                    max_output_size = 15,
                                                                    iou_threshold=0.5,
                                                                    score_threshold=float('-inf'),
                                                                    name=None)
                self.selected_boxes = tf.gather(boxes, self.selected_indices, axis=1)
                
                results = {} 
                results['detection_boxes'] = tf.concat([top_left_x, top_left_y, bottom_right_x, bottom_right_y], axis = -1)
                results['detection_classes'] = classes
                results['detection_scores'] = confidence
                return results
                '''
                boxes = tf.concat([top_left_x, top_left_y,
                            bottom_right_x, bottom_right_y,
                            confidence, classes], axis=-1)
                return boxes

            else: 
                selected_mask = tf.math.greater(confidence, tf.convert_to_tensor(np.array(self.threshold), tf.float32))
                total_num = tf.reduce_sum(tf.cast(selected_mask, tf.float32)) 
                
                #----------------------------------#
                # Remove some bbox with confidence lower than thd       
                #----------------------------------#
                confidence = tf.boolean_mask(confidence, selected_mask, axis=None)
                classes = tf.boolean_mask(classes, tf.squeeze(selected_mask, 2), axis=0)
                top_left_x = tf.boolean_mask(top_left_x, selected_mask, axis=None)
                top_left_y = tf.boolean_mask(top_left_y, selected_mask, axis=None)
                bottom_right_y = tf.boolean_mask(bottom_right_y, selected_mask, axis=None)
                bottom_right_x = tf.boolean_mask(bottom_right_x, selected_mask, axis=None)
                
                confidence = tf.reshape(tensor=confidence, shape=(-1, total_num, 1))
                classes = tf.reshape(tensor=classes, shape=(-1, total_num, self.num_class))
                top_left_x = tf.reshape(tensor=top_left_x, shape=(-1, total_num, 1))
                top_left_y = tf.reshape(tensor=top_left_y, shape=(-1, total_num, 1))
                bottom_right_x = tf.reshape(tensor=bottom_right_x, shape=(-1, total_num, 1))
                bottom_right_y = tf.reshape(tensor=bottom_right_y, shape=(-1, total_num, 1))

                for idx, i in enumerate(additional_class): 
                    tmp = tf.boolean_mask(i, tf.squeeze(selected_mask, 2), axis=0)
                    additional_class[idx] = tf.reshape(tensor=tmp, shape=(-1, total_num, list(self.multilabel_dict.values())[idx]))

                boxes = tf.concat([top_left_x, top_left_y,
                            bottom_right_x, bottom_right_y,
                            confidence, classes] + additional_class, axis=-1)

                return boxes
        #--------------------------------------------------#

        #----------------------------------#
        #     Network Type Choiches        #
        #----------------------------------#
        filters = {}
        if network_type == 'big':
            filters['a'] = 512
            filters['b'] = 256
            filters['c'] = 128
        elif network_type == 'medium':
            filters['a'] = 512
            filters['b'] = 128
            filters['c'] = 128
        elif network_type == 'small':
            filters['a'] = 256
            filters['b'] = 128
            filters['c'] = 128
        elif network_type == 'very_small':
            filters['a'] = 128
            filters['b'] = 64
            filters['c'] = 64
        elif network_type == 'special':
            filters['a'] = 64
            filters['b'] = 32
            filters['c'] = 32
        #---------------------------------#
        if network_type == 'special' or network_type == 'very_small':
            route1, route2, inputs, self.yolo_special_vars = darknet53(inputs, 
                                                                     training=is_training,
                                                                     data_format=data_format, 
                                                                     network_type=network_type)
            self.yolo_very_small_vars = self.yolo_special_vars
        else:
            route1, route2, inputs, _ = darknet53(inputs, 
                                                  training=is_training,
                                                  data_format=data_format, 
                                                  network_type=network_type)

        #-------------------------------------#
        #     get yolo small variables        #
        #-------------------------------------#
        self.yolo_small_vars = tf.global_variables(scope='yolo_v3_model')
        route, inputs = yolo_convolution_block(inputs, 
                                                filters=filters['a'], 
                                                training=is_training,
                                                data_format=data_format)
        inputs_detect1 = inputs

        inputs = new_conv2d_layer(input=route, 
                    filter_shape=[1, 1, route.get_shape().as_list()[-1], 256], 
                    name = 'main_input_conv13', 
                    dropout_val= self.dropout_val, 
                    activation = 'LRELU',
                    lrelu_alpha=self.leaky_relu_alpha,
                    padding='SAME', 
                    strides=[1, 1, 1, 1],
                    data_type=tf.float32,  
                    is_training=is_training,
                    use_bias=False,
                    use_batchnorm=True)

        upsample_size = route2.get_shape().as_list()
        inputs = upsample(inputs, 
                            out_shape=upsample_size,
                            data_format=data_format)
        axis = 3
        inputs = tf.concat([inputs, route2], axis=axis)

        #-------------------------------------#
        #     get yolo medium variables       #
        #-------------------------------------#
        self.yolo_medium_vars = tf.global_variables(scope='yolo_v3_model')
        route, inputs = yolo_convolution_block(inputs, 
                                                filters=filters['b'],  
                                                training=is_training,
                                                data_format=data_format)
        inputs_detect2 = inputs
        
        inputs = new_conv2d_layer(input=route, 
                                    filter_shape=[1, 1, route.get_shape().as_list()[-1], 128], 
                                    name = 'main_input_conv14', 
                                    dropout_val= self.dropout_val, 
                                    activation = 'LRELU',
                                    lrelu_alpha=self.leaky_relu_alpha,
                                    padding='SAME', 
                                    strides=[1, 1, 1, 1],
                                    data_type=tf.float32,  
                                    is_training=is_training,
                                    use_bias=False,
                                    use_batchnorm=True)
        
        upsample_size = route1.get_shape().as_list()
        inputs = upsample(inputs, 
                            out_shape=upsample_size,
                            data_format=data_format)
        inputs = tf.concat([inputs, route1], axis=axis)
        route, inputs = yolo_convolution_block(inputs, 
                                                filters=filters['c'], 
                                                training=is_training,
                                                data_format=data_format)
        inputs_detect3 = inputs

        #-------------------------------------#
        #       get yolo big variables        #
        #       get yolo all variables        #
        #-------------------------------------#
        self.yolo_big_vars = tf.global_variables(scope='yolo_v3_model')
        self.yolo_vars = tf.global_variables(scope='yolo_v3_model')


        if self.is_multilabel: 
            total_properties = 6 #sum(self.multilabel_dict.values)
            OUTPUT_DEPTH = int(len(self.anchor)/3 * (5 + self.num_class + total_properties))
        else: 
            OUTPUT_DEPTH = int(len(self.anchor)/3 * (5 + self.num_class))

        """
        self.detect1 = tf.layers.conv2d(inputs_detect1, 
                                    filters=OUTPUT_DEPTH,
                                    kernel_size=1, 
                                    strides=1, 
                                    use_bias=True,
                                    data_format=data_format)
     
        self.detect2 = tf.layers.conv2d(inputs_detect2, 
                                    filters=OUTPUT_DEPTH,
                                    kernel_size=1, 
                                    strides=1, 
                                    use_bias=True,
                                    data_format=data_format)

        self.detect3 = tf.layers.conv2d(inputs_detect3, 
                                    filters=OUTPUT_DEPTH,
                                    kernel_size=1, 
                                    strides=1, 
                                    use_bias=True,
                                    data_format=data_format)
        """
        self.detect1 = new_conv2d_layer(input=inputs_detect1, 
                                    filter_shape=[1, 1, inputs_detect1.get_shape().as_list()[-1], OUTPUT_DEPTH], 
                                    name = 'detect1_conv', 
                                    dropout_val= self.dropout_val, 
                                    activation = 'None',
                                    padding='VALID', 
                                    strides=[1, 1, 1, 1],
                                    is_training=is_training,
                                    use_bias=True,
                                    use_batchnorm=False)
        self.detect2 = new_conv2d_layer(input=inputs_detect2, 
                                    filter_shape=[1, 1, inputs_detect2.get_shape().as_list()[-1], OUTPUT_DEPTH], 
                                    name = 'detect2_conv', 
                                    dropout_val= self.dropout_val, 
                                    activation = 'None',
                                    padding='VALID', 
                                    strides=[1, 1, 1, 1],
                                    is_training=is_training,
                                    use_bias=True,
                                    use_batchnorm=False)
        self.detect3 = new_conv2d_layer(input=inputs_detect3, 
                                    filter_shape=[1, 1, inputs_detect3.get_shape().as_list()[-1], OUTPUT_DEPTH], 
                                    name = 'detect2_conv', 
                                    dropout_val= self.dropout_val, 
                                    activation = 'None',
                                    padding='VALID', 
                                    strides=[1, 1, 1, 1],
                                    is_training=is_training,
                                    use_bias=True,
                                    use_batchnorm=False)
        self.output_list = [self.detect1, self.detect2, self.detect3]

        combine_box1 = yolo_layer(self.detect1, 
                                n_classes=self.num_class,
                                anchors=self.anchor[6:9],
                                img_size=model_size,
                                data_format=data_format)
        combine_box2 = yolo_layer(self.detect2, 
                                n_classes=self.num_class,
                                anchors=self.anchor[3:6],
                                img_size=model_size,
                                data_format=data_format)
        combine_box3 = yolo_layer(self.detect3, 
                                n_classes=self.num_class,
                                anchors=self.anchor[0:3],
                                img_size=model_size,
                                data_format=data_format)
        
        if self.lite: 
            inputs = combine_box1
        else: 
            inputs = tf.concat([combine_box1, combine_box2, combine_box3], axis=1)
        self.boxes_dicts = build_boxes(inputs)
















import json
import cv2
import random
import numpy as np
import tensorflow as tf
import qoalai.networks.densenet as densenet
from simple_tensor.tensor_operations import *
from comdutils.file_utils import *
slim = tf.contrib.slim



class Landmark(object):
    def __init__(self,
                 num_landmark_point,
                 input_height = 400,
                 input_width = 400, 
                 input_channel = 3):

        """Constructor
        
        Arguments:
            classes {list of string} -- the image classes list

        Keyword Arguments:
            input_height {int} -- the height of input image (default: {512})
            input_width {int} -- the width of input image (default: {512})
            input_channel {int} -- the channel of input image (default: {3})
        """
        self.num_landmark_point = num_landmark_point
        self.input_height = input_height
        self.input_width = input_width
        self.input_channel = input_channel

        self.input_placeholder = tf.placeholder(tf.float32, shape=(None, self.input_height, self.input_width, self.input_channel))
        self.output_placeholder = tf.placeholder(tf.float32, shape=(None, 1, self.num_landmark_point, 2))

    
    def build_densenet_base(self, input_tensor,
                            dropout_rate,
                            is_training,
                            top_layer_depth):
        """[summary]
        
        Arguments:
            input_tensordropout_rate {[type]} -- [description]
            is_training {bool} -- [description]
        """
        arg_scoope = densenet.densenet_arg_scope()
        out = None

        with slim.arg_scope(arg_scoope):
            out = densenet.densenet121(inputs=input_tensor, 
                                       num_classes=1, 
                                       is_training=is_training)
            base_var_list = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)


        h = out.get_shape().as_list()[1]
        w = out.get_shape().as_list()[2]
        print ('-------------INFO-------------------')
        print ('INFO the height of the basenet output ' + str(h))
        print ('INFO the width of the basenet output ' + str(w))
        print ('------------------------------------')

        if w < self.num_landmark_point:
            print ('-------------WARNING---------------')
            print ('WARNING the number of your landmark point is to many')
            print ('WARNING the maximum number is: ' + str(w))
            print ('-----------------------------------')

        while(True):
            if w == self.num_landmark_point:
                break

            out = new_conv2d_layer(out, 
                                    filter_shape=[2, 2, out.get_shape().as_list()[-1], top_layer_depth], 
                                    name='cv1', 
                                    dropout_val=0.85, 
                                    activation = 'LRELU', 
                                    lrelu_alpha=0.2,
                                    padding='VALID', 
                                    strides=[1, 1, 1, 1],
                                    data_type=tf.float32,  
                                    is_training=is_training,
                                    use_bias=True,
                                    use_batchnorm=True) 
            w = out.get_shape().as_list()[2]

        while(True):
            if h == 1:
                break
            
            out = new_conv2d_layer(out, 
                                    filter_shape=[2, 2, out.get_shape().as_list()[-1], top_layer_depth], 
                                    name='cv1', 
                                    dropout_val=0.85, 
                                    activation = 'LRELU', 
                                    lrelu_alpha=0.2,
                                    padding='SAME', 
                                    strides=[1, 2, 1, 1],
                                    data_type=tf.float32,  
                                    is_training=is_training,
                                    use_bias=True,
                                    use_batchnorm=True) 
            h = out.get_shape().as_list()[1]

        out = new_conv2d_layer(out, 
                                    filter_shape=[1, 1, out.get_shape().as_list()[-1], 2], 
                                    name='cv1', 
                                    dropout_val=0.85, 
                                    activation = 'LRELU', 
                                    lrelu_alpha=0.2,
                                    padding='SAME', 
                                    strides=[1, 1, 1, 1],
                                    data_type=tf.float32,  
                                    is_training=is_training,
                                    use_bias=False,
                                    use_batchnorm=False) 
        out = tf.nn.sigmoid(out)
        print ('--------------INFO-----------------')
        print ('The shape of the detector is: (None,' + str(h) +' ' + str(w) + ' 2)')
        print ('-----------------------------------')
        return out


    def read_target(self, file_path):
        tmp = np.zeros((1, self.num_landmark_point,  2))

        #----------------------------------------------------------------#
        # this part is reading the label in a .txt file for a single image #
        #----------------------------------------------------------------#
        file = open(file_path, "r") 
        data = file.read()
        data = data.split()
        length = len(data)
        
        if length != 2 * self.num_landmark_point:
            print ("-------------------WARNING-------------------")
            print ("WRONG label format")
            print ("the label format is x1, y1, x2, y2, ..., xn, yn")
            print ("the value of x and y are relative to image width and height")
            print ("each image has single .txt file as label with the name")
            print ("---------------------------------------------")

        for i in range(self.num_landmark_point):
            tmp[0, i, 0] = float(data[i*2])
            tmp[0, i, 1] = float(data[i*2 + 1])

        return tmp


    def batch_generator(self, batch_size, dataset_path, message):
        """Train Generator
        
        Arguments:
            batch_size {integer} -- the size of the batch
            image_name_list {list of string} -- the list of image name
        """
        label_folder_path = dataset_path + "labels/"
        dataset_folder_path = dataset_path + "images/"
        dataset_file_list = get_filenames(dataset_folder_path)
        random.shuffle(dataset_file_list)
        
        print ("------------------------INFO-------------------")
        print ("Image Folder: " + dataset_folder_path)
        print ("Number of Image: " + str(len(dataset_file_list)))
        print ("-----------------------------------------------")

        # Infinite loop.
        idx = 0
        while True:
            x_batch = []
            y_pred = []

            for i in range(batch_size):
                if idx >= len(dataset_file_list):
                    random.shuffle(dataset_file_list)
                    print ("==>>> INFO: your " + message +" dataset is reshuffled again: ", idx)
                    idx = 0
                try:
                    tmp_x = cv2.imread(dataset_folder_path + dataset_file_list[idx])
                    tmp_x = cv2.cvtColor(tmp_x, cv2.COLOR_BGR2RGB)
                    tmp_x = cv2.resize(tmp_x, (self.input_width, self.input_height))
                    tmp_x = tmp_x.astype(np.float32) / 255.
                    tmp_y = self.read_target(label_folder_path + dataset_file_list[idx].split('.')[0] + ".txt")
                    x_batch.append(tmp_x)
                    y_pred.append(tmp_y)
                except Exception as e:
                    print ("---------------------------------------------------------------")
                    print ("WARNING: the " + str(e))
                    print ("---------------------------------------------------------------")

                idx += 1
            yield (np.array(x_batch), np.array(y_pred))


    def optimize(self, 
                 iteration, 
                 subdivition,
                 cost_tensor,
                 optimizer_tensor,
                 out_tensor, 
                 session,
                 saver, 
                 train_generator,
                 val_generator,
                 best_loss,
                 path_tosave_model='model/model1'):
        """[summary]
        
        Arguments:
            iteration {[type]} -- [description]
            subdivition {[type]} -- [description]
            cost_tensor {[type]} -- [description]
            optimizer_tensor {[type]} -- [description]
            out_tensor {[type]} -- [description]
        
        Keyword Arguments:
            train_batch_size {int} -- [description] (default: {32})
            val_batch_size {int} -- [description] (default: {50})
            path_tosave_model {str} -- [description] (default: {'model/model1'})
        """

        self.train_loss = []
        self.val_loss = []
        best_loss = best_loss
        
        for i in range(iteration):
            sign = "-"
            t_losses = []

            for j in range(subdivition):
                x_train, y_train = next(train_generator)
                feed_dict = {}
                feed_dict[self.input_placeholder] = x_train
                feed_dict[self.output_placeholder] = y_train
                session.run(optimizer_tensor, feed_dict)
                loss = session.run(cost_tensor, feed_dict)
                t_losses.append(loss)
                print ("> Train sub", j, 'loss : ', loss)
                
            x_val, y_val = next(val_generator)
            feed_dict = {}
            feed_dict[self.input_placeholder] = x_val
            feed_dict[self.output_placeholder] = y_val
            loss = session.run(cost_tensor, feed_dict)
            
            t_loss = sum(t_losses) / (len(t_losses) + 0.0001)
            self.train_loss.append(t_loss)
            self.val_loss.append(loss)
                
            if best_loss > loss:
                best_loss = loss
                sign = "************* model saved"
                saver.save(session, path_tosave_model)
        
            print (">> epoch: ", i, "train loss: ", round(t_loss, 3), "val loss: ", round(loss, 3), sign)    